__author__ = "Simon Nilsson"

import os

import numpy as np
import pandas as pd
from shapely import geometry
from shapely.geometry import Point, Polygon

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_if_filepath_list_is_empty,
                                check_that_column_exist)
from simba.utils.data import detect_bouts
from simba.utils.errors import NoChoosenClassifierError, NoROIDataError
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_config_entry, read_df
from simba.utils.warnings import NoDataFoundWarning, ROIWarning


class ROIClfCalculator(ConfigReader):
    """
    Compute aggregate statistics of classification results within user-defined ROIs.
    Results are stored in `project_folder/logs` directory of the SimBA project.

    :param str config_path: path to SimBA project config file in Configparser format

    .. note:
       'GitHub tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-4--analyze-machine-results`__.

    Examples
    -----
    >>> clf_ROI_analyzer = ROIClfCalculator(config_ini="MyConfigPath")
    >>> clf_ROI_analyzer.run(behavior_list=['Attack', 'Sniffing'], ROI_dict_lists={'Rectangle': ['rec'], 'Circle': ['Stimulus 1', 'Stimulus 2', 'Stimulus 3']}, body_part_list=['Nose_1'], measurements=['Total time by ROI (s)', 'Started bouts by ROI (count)', 'Ended bouts by ROI (count)'])
    """

    def __init__(self, config_ini: str):
        ConfigReader.__init__(self, config_path=config_ini)
        self.read_roi_data()

    def __inside_rectangle(
        self, bp_x, bp_y, top_left_x, top_left_y, bottom_right_x, bottom_right_y
    ):
        """
        Private helper to calculate if body-part is inside a rectangle.
        """
        if ((top_left_x) <= bp_x <= (bottom_right_x)) and (
            (top_left_y) <= bp_y <= (bottom_right_y)
        ):
            return 1
        else:
            return 0

    def __inside_circle(self, bp_x, bp_y, center_x, center_y, radius):
        """
        Private helper to calculate if body-part is inside a circle.
        """
        px_dist = int(np.sqrt((bp_x - center_x) ** 2 + (bp_y - center_y) ** 2))
        if px_dist <= radius:
            return 1
        else:
            return 0

    def __inside_polygon(self, bp_x, bp_y, polygon):
        """
        Private helper to calculate if body-part is inside a polygon.
        """
        if polygon.contains(Point(int(bp_x), int(bp_y))):
            return 1
        else:
            return 0

    def __compute_agg_statistics(self, data: pd.DataFrame):
        """

        Parameters
        ----------
        data: pd.DataFrame
            Dataframe with boolean columns representing behaviors (behavior present: 1, behavior absent: 0)
            and ROI data (inside ROI: 1, outside ROI: 0).

        """
        self.results_dict[self.video_name] = {}
        for clf in self.behavior_list:
            self.results_dict[self.video_name][clf] = {}
            for roi in self.found_rois:
                self.results_dict[self.video_name][clf][roi] = {}
                if "Total time by ROI (s)" in self.measurements:
                    frame_cnt = len(data.loc[(data[clf] == 1) & (data[roi] == 1)])
                    if frame_cnt > 0:
                        self.results_dict[self.video_name][clf][roi][
                            "Total time by ROI (s)"
                        ] = (frame_cnt / self.fps)
                    else:
                        self.results_dict[self.video_name][clf][roi][
                            "Total time (s)"
                        ] = 0
                if "Started bouts by ROI (count)" in self.measurements:
                    start_frames = list(
                        detect_bouts(data_df=data, target_lst=[clf], fps=int(self.fps))[
                            "Start_frame"
                        ]
                    )
                    self.results_dict[self.video_name][clf][roi][
                        "Started bouts by ROI (count)"
                    ] = len(data[(data.index.isin(start_frames)) & (data[roi] == 1)])
                if "Ended bouts by ROI (count)" in self.measurements:
                    start_frames = list(
                        detect_bouts(data_df=data, target_lst=[clf], fps=int(self.fps))[
                            "End_frame"
                        ]
                    )
                    self.results_dict[self.video_name][clf][roi][
                        "Ended bouts by ROI (count)"
                    ] = len(data[(data.index.isin(start_frames)) & (data[roi] == 1)])

    def __print_missing_roi_warning(self, roi_type: str, roi_name: str):
        """
        Private helper to print warnings when ROI shapes have been defined in some videos but missing in others.
        """
        names = "None"
        ROIWarning(
            msg=f'ROI named "{roi_name}" of shape type "{roi_type}" not found for video {self.video_name}. Skipping shape...'
        )
        if roi_type.lower() == "rectangle":
            names = list(
                self.rectangles_df["Name"][
                    self.rectangles_df["Video"] == self.video_name
                ]
            )
        elif roi_type.lower() == "circle":
            names = list(
                self.circles_df["Name"][self.circles_df["Video"] == self.video_name]
            )
        elif roi_type.lower() == "polygon":
            names = list(
                self.polygon_df["Name"][self.polygon_df["Video"] == self.video_name]
            )
        ROIWarning(
            msg=f"NOTE: Video {self.video_name} has the following {roi_type} shape names: {names}"
        )

    def run(
        self,
        ROI_dict_lists: dict,
        measurements: list,
        behavior_list: list,
        body_part_list: list,
    ):
        """
        Parameters
        ----------
        ROI_dict_lists: dict
            A dictionary with the shape type as keys (i.e., Rectangle, Circle, Polygon) and lists of shape names
            as values.
        measurements: list
            Measurements to calculate aggregate statistics for. E.g., ['Total time by ROI (s)', 'Started bouts', 'Ended bouts']
        behavior_list: list
            Classifier names to calculate ROI statistics. E.g., ['Attack', 'Sniffing']
        body_part_list: list
            Body-part names to use to infer animal location. Eg., ['Nose_1'].
        """

        self.ROI_dict_lists, self.behavior_list, self.measurements = (
            ROI_dict_lists,
            self.clf_names,
            measurements,
        )
        self.file_type = read_config_entry(
            config=self.config,
            section="General settings",
            option="workflow_file_type",
            data_type="str",
        )
        check_if_filepath_list_is_empty(
            filepaths=self.machine_results_paths,
            error_msg="SIMBA ERROR: No machine learning results found in the project_folder/csv/machine_results directory. Create machine classifications before analyzing classifications by ROI",
        )
        if len(behavior_list) == 0:
            raise NoChoosenClassifierError()
        print(f"Analyzing {str(len(self.machine_results_paths))} files...")
        body_part_col_names = []
        body_part_col_names_x, body_part_col_names_y = [], []
        for body_part in body_part_list:
            body_part_col_names.extend(
                (body_part + "_x", body_part + "_y", body_part + "_p")
            )
            body_part_col_names_x.append(body_part + "_x")
            body_part_col_names_y.append(body_part + "_y")
        all_columns = body_part_col_names + self.behavior_list
        self.results_dict = {}

        self.frame_counter_dict = {}
        for file_cnt, file_path in enumerate(self.machine_results_paths):
            _, self.video_name, ext = get_fn_ext(file_path)
            print("Analyzing {}....".format(self.video_name))
            data_df = read_df(file_path, self.file_type)
            for column in all_columns:
                check_that_column_exist(
                    file_name=self.video_name, df=data_df, column_name=column
                )
            data_df = data_df[all_columns]
            self.results = data_df[self.behavior_list]
            shapes_in_video = (
                len(
                    self.rectangles_df.loc[
                        (self.rectangles_df["Video"] == self.video_name)
                    ]
                )
                + len(
                    self.circles_df.loc[(self.circles_df["Video"] == self.video_name)]
                )
                + len(
                    self.polygon_df.loc[(self.polygon_df["Video"] == self.video_name)]
                )
            )
            if shapes_in_video == 0:
                NoDataFoundWarning(
                    msg="Skipping {self.video_name}: Video {self.video_name} has 0 user-defined ROI shapes."
                )
                continue
            _, _, self.fps = self.read_video_info(video_name=self.video_name)
            self.found_rois = []
            for roi_type, roi_data in self.ROI_dict_lists.items():
                shape_info = pd.DataFrame()
                for roi_name in roi_data:
                    if roi_type.lower() == "rectangle":
                        shape_info = self.rectangles_df.loc[
                            (self.rectangles_df["Video"] == self.video_name)
                            & (self.rectangles_df["Shape_type"] == roi_type)
                            & (self.rectangles_df["Name"] == roi_name)
                        ]
                    elif roi_type.lower() == "circle":
                        shape_info = self.circles_df.loc[
                            (self.circles_df["Video"] == self.video_name)
                            & (self.circles_df["Shape_type"] == roi_type)
                            & (self.circles_df["Name"] == roi_name)
                        ]
                    elif roi_type.lower() == "polygon":
                        shape_info = self.polygon_df.loc[
                            (self.polygon_df["Video"] == self.video_name)
                            & (self.polygon_df["Shape_type"] == roi_type)
                            & (self.polygon_df["Name"] == roi_name)
                        ]
                    if len(shape_info) == 0:
                        self.__print_missing_roi_warning(
                            roi_type=roi_type, roi_name=roi_name
                        )
                        continue
                    if roi_type.lower() == "rectangle":
                        data_df["top_left_x"], data_df["top_left_y"] = (
                            shape_info["topLeftX"].values[0],
                            shape_info["topLeftY"].values[0],
                        )
                        data_df["bottom_right_x"], data_df["bottom_right_y"] = (
                            shape_info["Bottom_right_X"].values[0],
                            shape_info["Bottom_right_Y"].values[0],
                        )
                        self.results[roi_name] = data_df.apply(
                            lambda x: self.__inside_rectangle(
                                bp_x=x[body_part_col_names_x[0]],
                                bp_y=x[body_part_col_names_y[0]],
                                top_left_x=x["top_left_x"],
                                top_left_y=x["top_left_y"],
                                bottom_right_x=x["bottom_right_x"],
                                bottom_right_y=x["bottom_right_y"],
                            ),
                            axis=1,
                        )
                        self.found_rois.append(roi_name)
                    elif roi_type.lower() == "circle":
                        data_df["center_x"], data_df["center_y"], data_df["radius"] = (
                            shape_info["centerX"].values[0],
                            shape_info["centerY"].values[0],
                            shape_info["radius"].values[0],
                        )
                        self.results[roi_name] = data_df.apply(
                            lambda x: self.__inside_circle(
                                bp_x=x[body_part_col_names_x[0]],
                                bp_y=x[body_part_col_names_y[0]],
                                center_x=x["center_x"],
                                center_y=x["center_y"],
                                radius=x["radius"],
                            ),
                            axis=1,
                        )
                        self.found_rois.append(roi_name)
                    elif roi_type.lower() == "polygon":
                        polygon_vertices = []
                        for i in shape_info["vertices"].values[0]:
                            polygon_vertices.append(geometry.Point(i))
                        polygon = Polygon([[p.x, p.y] for p in polygon_vertices])
                        self.results[roi_name] = data_df.apply(
                            lambda x: self.__inside_polygon(
                                bp_x=x[body_part_col_names_x[0]],
                                bp_y=x[body_part_col_names_y[0]],
                                polygon=polygon,
                            ),
                            axis=1,
                        )
                        self.found_rois.append(roi_name)
                self.__compute_agg_statistics(data=self.results)
        self.__organize_output_data()

    def __organize_output_data(self):
        """
        Helper to organize the results[dict] into a human-readable CSV file.
        """
        if len(self.results_dict.keys()) == 0:
            raise NoROIDataError(
                msg="ZERO ROIs found the videos represented in the project_folder/csv/machine_results directory"
            )
        out_df = pd.DataFrame(
            columns=["VIDEO", "CLASSIFIER", "ROI", "MEASUREMENT", "VALUE"]
        )
        for video_name, video_data in self.results_dict.items():
            for clf, clf_data in video_data.items():
                for roi_name, roi_data in clf_data.items():
                    for measurement_name, mesurement_value in roi_data.items():
                        out_df.loc[len(out_df)] = [
                            video_name,
                            clf,
                            roi_name,
                            measurement_name,
                            mesurement_value,
                        ]
        out_path = os.path.join(
            self.logs_path, f"Classification_time_by_ROI_{self.datetime}.csv"
        )
        out_df.to_csv(out_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Classification data by ROIs saved in {out_path}.",
            elapsed_time=self.timer.elapsed_time_str,
        )


#
# clf_ROI_analyzer = clf_within_ROI(config_ini="/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini")
# clf_ROI_analyzer.run(behavior_list=['Attack', 'Sniffing'], ROI_dict_lists={'Rectangle': ['rec'], 'Circle': ['Stimulus 1', 'Stimulus 2', 'Stimulus 3']}, body_part_list=['Nose_1'], measurements=['Total time by ROI (s)', 'Started bouts by ROI (count)', 'Ended bouts by ROI (count)'])
#

# test = ROIClfCalculator(config_ini="/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini")
# test.run(behavior_list=['Attack', 'Sniffing'], ROI_dict_lists={'Rectangle': ['DAMN'], 'Circle': [], 'Polygon': ['YOU_SUCK_SIMON']}, body_part_list=['Nose_1'], measurements=['Total time by ROI (s)', 'Started bouts by ROI (count)', 'Ended bouts by ROI (count)'])



import itertools
import os
from typing import List

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.utils.enums import DirNames
from simba.utils.errors import ROICoordinatesNotFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


class ROITimebinCalculator(ConfigReader):
    """
    Calculate how much time and how many entries animals are making into user-defined ROIs
    within user-defined time bins. Results are stored in the ``project_folder/logs`` directory of
    the SimBA project.

    :param str config_path: path to SimBA project config file in Configparser format
    :param int bin_length: length of time bins in seconds.

    Examples
    ----------
    >>> roi_time_bin_calculator = ROITimebinCalculator(config_path='MySimBaConfigPath', bin_length=15, body_parts=['Nose_1'], threshold=0.00)
    >>> roi_time_bin_calculator.run()
    >>> roi_time_bin_calculator.save()
    """

    def __init__(
        self, config_path: str, bin_length: int, body_parts: List[str], threshold: float
    ):
        ConfigReader.__init__(self, config_path=config_path)
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(
                expected_file_path=self.roi_coordinates_path
            )
        self.read_roi_data()
        self.bin_length, self.body_parts, self.threshold = (
            bin_length,
            body_parts,
            threshold,
        )
        self.save_path_time = os.path.join(
            self.logs_path, f"ROI_time_bins_{bin_length}s_time_data_{self.datetime}.csv"
        )
        self.save_path_entries = os.path.join(
            self.logs_path,
            f"ROI_time_bins_{bin_length}s_entry_data_{self.datetime}.csv",
        )
        settings = {"threshold": threshold, "body_parts": {}}
        for i in body_parts:
            animal_name = self.find_animal_name_from_body_part_name(
                bp_name=i, bp_dict=self.animal_bp_dict
            )
            settings["body_parts"][animal_name] = i
        self.roi_analyzer = ROIAnalyzer(
            ini_path=self.config_path,
            data_path=DirNames.OUTLIER_MOVEMENT_LOCATION.value,
            calculate_distances=False,
            settings=settings,
        )

        self.roi_analyzer.run()
        self.animal_names = list(self.roi_analyzer.bp_dict.keys())
        self.entries_exits_df = self.roi_analyzer.detailed_df

    #
    def run(self):
        self.out_time_lst, self.out_entries_lst = [], []
        print(
            f"Analyzing time-bin data for {len(self.outlier_corrected_paths)} video(s)..."
        )
        for file_cnt, file_path in enumerate(self.outlier_corrected_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(filepath=file_path)
            _, _, fps = self.read_video_info(video_name=self.video_name)
            frames_per_bin = int(fps * self.bin_length)
            video_frms = list(
                range(0, len(read_df(file_path=file_path, file_type=self.file_type)))
            )
            frame_bins = [
                video_frms[i * frames_per_bin : (i + 1) * frames_per_bin]
                for i in range((len(video_frms) + frames_per_bin - 1) // frames_per_bin)
            ]
            self.video_data = self.entries_exits_df[
                self.entries_exits_df["VIDEO"] == self.video_name
            ]
            for animal_name, shape_name in list(
                itertools.product(self.animal_names, self.shape_names)
            ):
                results_entries = pd.DataFrame(
                    columns=["VIDEO", "SHAPE", "ANIMAL", "TIME BIN", "ENTRY COUNT"]
                )
                results_time = pd.DataFrame(
                    columns=[
                        "VIDEO",
                        "SHAPE",
                        "ANIMAL",
                        "TIME BIN",
                        "TIME INSIDE SHAPE (S)",
                    ]
                )
                data_df = self.video_data.loc[
                    (self.video_data["SHAPE"] == shape_name)
                    & (self.video_data["ANIMAL"] == animal_name)
                ]
                entry_frms = list(data_df["ENTRY FRAMES"])
                inside_shape_frms = [
                    list(range(x, y))
                    for x, y in zip(
                        list(data_df["ENTRY FRAMES"].astype(int)),
                        list(data_df["EXIT FRAMES"].astype(int) + 1),
                    )
                ]
                inside_shape_frms = [i for s in inside_shape_frms for i in s]
                for bin_cnt, bin_frms in enumerate(frame_bins):
                    frms_inside_roi_in_timebin = [
                        x for x in inside_shape_frms if x in bin_frms
                    ]
                    entry_roi_in_timebin = [x for x in entry_frms if x in bin_frms]
                    results_time.loc[len(results_time)] = [
                        self.video_name,
                        shape_name,
                        animal_name,
                        bin_cnt,
                        len(frms_inside_roi_in_timebin) / fps,
                    ]
                    results_entries.loc[len(results_entries)] = [
                        self.video_name,
                        shape_name,
                        animal_name,
                        bin_cnt,
                        len(entry_roi_in_timebin),
                    ]
                self.out_time_lst.append(results_time)
                self.out_entries_lst.append(results_entries)
            video_timer.stop_timer()
            print(
                f"Video {self.video_name} complete (elapsed time {video_timer.elapsed_time_str}s)"
            )
        self.out_time = pd.concat(self.out_time_lst, axis=0).sort_values(
            by=["VIDEO", "SHAPE", "ANIMAL", "TIME BIN"]
        )
        self.out_entries = pd.concat(self.out_entries_lst, axis=0).sort_values(
            by=["VIDEO", "SHAPE", "ANIMAL", "TIME BIN"]
        )

    def save(self):
        self.out_entries.to_csv(self.save_path_entries)
        self.out_time.to_csv(self.save_path_time)
        self.timer.stop_timer()
        stdout_success(
            msg=f"ROI time bin entry data saved at {self.save_path_entries}",
            elapsed_time=self.timer.elapsed_time_str,
        )
        stdout_success(
            msg=f"ROI time bin time data saved at {self.save_path_time}",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = ROITimebinCalculator(config_path=r"/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini",
#                             bin_length=5, body_parts=['Simon_Nose_1_1', 'JJ_Nose_1_2'], threshold=0.00)
# test.run()
# test.save()



import glob
import os

import pandas as pd

from simba.utils.enums import ConfigKey, Keys, Paths
from simba.utils.errors import NoROIDataError
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_config_file


def create_emty_df(shape_type):
    col_list = None
    if shape_type == Keys.ROI_RECTANGLES.value:
        col_list = [
            "Video",
            "Shape_type",
            "Name",
            "Color name",
            "Color BGR",
            "Thickness",
            "topLeftX",
            "topLeftY",
            "Bottom_right_X",
            "Bottom_right_Y",
            "width",
            "height",
            "Tags",
            "Ear_tag_size",
        ]
    if shape_type == Keys.ROI_CIRCLES.value:
        col_list = [
            "Video",
            "Shape_type",
            "Name",
            "Color name",
            "Color BGR",
            "Thickness",
            "centerX",
            "centerY",
            "radius",
            "Tags",
            "Ear_tag_size",
        ]
    if shape_type == Keys.ROI_POLYGONS.value:
        col_list = [
            "Video",
            "Shape_type",
            "Name",
            "Color name",
            "Color BGR",
            "Thickness",
            "Center_X",
            "Center_Y",
            "vertices",
            "Tags",
            "Ear_tag_size",
        ]
    return pd.DataFrame(columns=col_list)


def multiply_ROIs(config_path, filename):
    _, CurrVidName, ext = get_fn_ext(filename)
    config = read_config_file(config_path=config_path)
    projectPath = config.get(
        ConfigKey.GENERAL_SETTINGS.value, ConfigKey.PROJECT_PATH.value
    )
    videoPath = os.path.join(projectPath, "videos")
    ROIcoordinatesPath = os.path.join(projectPath, "logs", Paths.ROI_DEFINITIONS.value)

    if not os.path.isfile(ROIcoordinatesPath):
        raise NoROIDataError(
            msg="Cannot multiply ROI definitions: no ROI definitions exist in SimBA project"
        )
    rectanglesInfo = pd.read_hdf(ROIcoordinatesPath, key=Keys.ROI_RECTANGLES.value)
    circleInfo = pd.read_hdf(ROIcoordinatesPath, key=Keys.ROI_CIRCLES.value)
    polygonInfo = pd.read_hdf(ROIcoordinatesPath, key=Keys.ROI_POLYGONS.value)

    try:
        r_df = rectanglesInfo[rectanglesInfo["Video"] == CurrVidName]
    except KeyError:
        r_df = create_emty_df("rectangle")

    try:
        c_df = circleInfo.loc[circleInfo["Video"] == str(CurrVidName)]
    except KeyError:
        c_df = create_emty_df("circle")

    try:
        p_df = polygonInfo.loc[polygonInfo["Video"] == str(CurrVidName)]
    except KeyError:
        p_df = create_emty_df("polygon")

    if len(r_df) == 0 and len(c_df) == 0 and len(p_df) == 0:
        print(
            "Cannot replicate ROIs to all videos: no ROI records exist for "
            + str(CurrVidName)
        )

    else:
        videofilesFound = glob.glob(videoPath + "/*.mp4") + glob.glob(
            videoPath + "/*.avi"
        )
        duplicatedRec, duplicatedCirc, duplicatedPoly = (
            r_df.copy(),
            c_df.copy(),
            p_df.copy(),
        )
        for vids in videofilesFound:
            _, vid_name, ext = get_fn_ext(vids)
            duplicatedRec["Video"], duplicatedCirc["Video"], duplicatedPoly["Video"] = (
                vid_name,
                vid_name,
                vid_name,
            )
            r_df = r_df.append(duplicatedRec, ignore_index=True)
            c_df = c_df.append(duplicatedCirc, ignore_index=True)
            p_df = p_df.append(duplicatedPoly, ignore_index=True)
        r_df = r_df.drop_duplicates(subset=["Video", "Name"], keep="first")
        c_df = c_df.drop_duplicates(subset=["Video", "Name"], keep="first")
        p_df = p_df.drop_duplicates(subset=["Video", "Name"], keep="first")

        store = pd.HDFStore(ROIcoordinatesPath, mode="w")
        store["rectangles"] = r_df
        store["circleDf"] = c_df
        store["polygons"] = p_df
        store.close()
        stdout_success(msg=f"ROI(s) for {CurrVidName} applied to all videos")
        print()
        print(
            'Next, click on "draw" to modify ROI location(s) or click on "reset" to remove ROI drawing(s)'
        )



__author__ = "Simon Nilsson"

import os

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import Formats, Paths
from simba.utils.printing import stdout_success
from simba.utils.read_write import (find_all_videos_in_directory,
                                    get_video_meta_data)
from simba.utils.warnings import NoFileFoundWarning

CLUSTER = "CLUSTER"
FPS = "fps"
VIDEO_SPEED = "video_speed"
START_FRAME = "START_FRAME"
POSE = "pose"
CREATE = "create"
CIRCLE_SIZE = "circle_size"


class ClusterVisualizer(ConfigReader, UnsupervisedMixin):
    """
    Class for creating video examples of cluster assignments.

    :param str config_path: path to SimBA configparser.ConfigParser project_config.ini
    :param str data_path: path to pickle holding unsupervised results in ``data_map.yaml`` format.
    :param str video_dir: path to directory holding videos.
    :param dict settings: dict holding attributes of the videos

    :example:
    >>> settings = {'video_speed': 0.5, 'pose': {'create': True, 'circle_size': 5}}
    >>> visualizer = ClusterVisualizer(video_dir='unsupervised/project_folder/videos', data_path='unsupervised/cluster_models/quizzical_rhodes.pickle', settings=settings, config_path='unsupervised/project_folder/project_config.ini')
    >>> visualizer.run()
    """

    def __init__(
        self, config_path: str, video_dir: str, data_path: str, settings: dict
    ):
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        self.settings, self.video_dir, self.data_path = settings, video_dir, data_path
        self.save_parent_dir = os.path.join(
            self.project_path, Paths.CLUSTER_EXAMPLES.value
        )
        if not os.path.exists(self.save_parent_dir):
            os.makedirs(self.save_parent_dir)
        check_file_exist_and_readable(file_path=data_path)
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.video_files = find_all_videos_in_directory(
            directory=video_dir, as_dict=True
        )

    def run(self):
        self.video_counter = 0
        self.data = self.read_pickle(data_path=self.data_path)
        self.cluster_data = self.data[Clustering.CLUSTER_MODEL.value][
            Unsupervised.MODEL.value
        ].labels_
        self.x_df = self.data[Unsupervised.METHODS.value][
            Unsupervised.SCALED_DATA.value
        ].reset_index()
        self.x_y_df = pd.concat(
            [self.x_df, pd.DataFrame(self.cluster_data, columns=[CLUSTER])], axis=1
        )
        self.bout_cnt = len(
            self.data[Unsupervised.DATA.value][Unsupervised.BOUTS_FEATURES.value]
        )
        for cluster_id in np.unique(self.cluster_data):
            self.cluster_id = cluster_id
            for video_name in self.data[Unsupervised.DATA.value][
                Unsupervised.BOUTS_FEATURES.value
            ][Unsupervised.VIDEO.value].unique():
                self.video_name = video_name
                if video_name not in self.video_files.keys():
                    NoFileFoundWarning(
                        msg=f"Video {video_name} not found in video directory {self.video_dir}"
                    )
                    continue
                else:
                    self.cluster_video_df = self.x_y_df.loc[
                        (self.x_y_df[Unsupervised.VIDEO.value] == video_name)
                        & (self.x_y_df[CLUSTER] == cluster_id)
                    ]
                    if len(self.cluster_video_df) > 0:
                        self.__create_videos()
        self.timer.stop_timer()
        stdout_success(
            msg=f"Visualizations complete. Data saved at {self.save_parent_dir}",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __create_videos(self):
        self.save_directory = os.path.join(
            self.save_parent_dir, str(self.cluster_id), self.video_name
        )
        if not os.path.exists(self.save_directory):
            os.makedirs(self.save_directory)
        video_meta_data = get_video_meta_data(
            video_path=self.video_files[
                self.cluster_video_df[Unsupervised.VIDEO.value].values[0]
            ]
        )
        output_fps = max(1, int(video_meta_data[FPS] * self.settings[VIDEO_SPEED]))
        cap = cv2.VideoCapture(
            self.video_files[self.cluster_video_df["VIDEO"].values[0]]
        )
        cluster_frames = list(
            self.cluster_video_df.apply(
                lambda x: list(
                    range(
                        int(x[Unsupervised.START_FRAME.value]),
                        int(x[Unsupervised.END_FRAME.value]) + 1,
                    )
                ),
                1,
            )
        )
        for cluster_event_cnt, cluster_event in enumerate(cluster_frames):
            save_path = os.path.join(
                self.save_directory, f"Event_{str(cluster_event_cnt)}.mp4"
            )
            self.writer = cv2.VideoWriter(
                save_path,
                self.fourcc,
                output_fps,
                (video_meta_data["width"], video_meta_data["height"]),
            )
            start_frm, end_frm, current_frm = (
                cluster_event[0],
                cluster_event[-1],
                cluster_event[0],
            )
            cluster_event_frms = end_frm - start_frm
            cap.set(1, cluster_event[0])
            frame_cnt = 0
            while current_frm < end_frm:
                _, img = cap.read()
                if self.settings[POSE][CREATE]:
                    bp_data = self.data[Unsupervised.DATA.value]["FRAME_POSE"].iloc[
                        current_frm
                    ]
                    for animal_cnt, (animal_name, animal_bps) in enumerate(
                        self.animal_bp_dict.items()
                    ):
                        for bp_cnt, bp in enumerate(
                            zip(animal_bps["X_bps"], animal_bps["Y_bps"])
                        ):
                            x_bp, y_bp = bp_data[bp[0]], bp_data[bp[1]]
                            cv2.circle(
                                img,
                                (int(x_bp), int(y_bp)),
                                self.settings[POSE][CIRCLE_SIZE],
                                self.clr_lst[animal_cnt][bp_cnt],
                                -1,
                            )
                self.writer.write(img)
                print(
                    f"Writing frame {str(frame_cnt)}/{str(cluster_event_frms)}, "
                    f"Bout {str(cluster_event_cnt+1)}/{len(cluster_frames)}, "
                    f""
                    f"Cluster: {self.cluster_id}, "
                    f"Video: {self.video_name}, "
                    f"Total bout count: {self.video_counter}/{self.bout_cnt}..."
                )
                current_frm += 1
                frame_cnt += 1
            cap.release()
            self.writer.release()
            self.video_counter += 1


# settings = {'video_speed': 0.5, 'pose': {'create': True, 'circle_size': 5}}
# test = ClusterVisualizer(video_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/videos',
#                          data_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models/quizzical_rhodes.pickle',
#                          settings=settings,
#                          config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini')
# test.run()



__author__ = "Simon Nilsson"

from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.third_party_label_appenders.third_party_appender import \
    ThirdPartyLabelAppender
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        FolderSelect)
from simba.utils.checks import check_if_dir_exists
from simba.utils.enums import Formats, Keys, Links, Options
from simba.utils.lookups import get_third_party_appender_file_formats


class ThirdPartyAnnotatorAppenderPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="APPEND THIRD-PARTY ANNOTATIONS")
        ConfigReader.__init__(self, config_path=config_path)
        apps_lst = Options.THIRD_PARTY_ANNOTATION_APPS_OPTIONS.value
        warnings_lst = Options.THIRD_PARTY_ANNOTATION_ERROR_OPTIONS.value
        app_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="THIRD-PARTY APPLICATION",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.THIRD_PARTY_ANNOTATION_NEW.value,
        )
        self.app_dropdown = DropDownMenu(
            app_frm, "THIRD-PARTY APPLICATION:", apps_lst, "35"
        )
        self.app_dropdown.setChoices(apps_lst[0])
        app_frm.grid(row=0, column=0, sticky=NW)
        self.app_dropdown.grid(row=0, column=0, sticky=NW)

        select_data_frm = LabelFrame(
            self.main_frm,
            text="SELECT DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.data_folder = FolderSelect(select_data_frm, "DATA DIRECTORY:", lblwidth=35)
        select_data_frm.grid(row=1, column=0, sticky=NW)
        self.data_folder.grid(row=0, column=0, sticky=NW)

        self.error_dropdown_dict = self.create_dropdown_frame(
            main_frm=self.main_frm,
            drop_down_titles=warnings_lst,
            drop_down_options=["WARNING", "ERROR"],
            frm_title="WARNINGS AND ERRORS",
        )
        log_frm = LabelFrame(
            self.main_frm, text="LOGGING", font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        self.log_var = BooleanVar(value=True)
        self.log_cb = Checkbutton(
            log_frm, text="CREATE IMPORT LOG", variable=self.log_var
        )
        log_frm.grid(row=5, column=0, sticky=NW)
        self.log_cb.grid(row=0, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

    def run(self):
        settings = {"log": self.log_var.get()}
        settings["file_format"] = get_third_party_appender_file_formats()[
            self.app_dropdown.getChoices()
        ]
        settings["errors"], app_choice = {}, self.app_dropdown.getChoices()
        for error_name, error_dropdown in self.error_dropdown_dict.items():
            settings["errors"][error_name] = error_dropdown.getChoices()
        check_if_dir_exists(in_dir=self.data_folder.folder_path)

        third_party_importer = ThirdPartyLabelAppender(
            app=self.app_dropdown.getChoices(),
            config_path=self.config_path,
            data_dir=self.data_folder.folder_path,
            settings=settings,
        )
        third_party_importer.run()



__author__ = "Simon Nilsson"

from typing import List

import pandas as pd
from joblib import Parallel, delayed
from joblib.externals.loky import get_reusable_executor

from simba.utils.data import detect_bouts
from simba.utils.read_write import read_video_info


def bout_aggregator(
    data: pd.DataFrame,
    clfs: List[str],
    feature_names: List[str],
    aggregator: str,
    min_bout_length: int,
    video_info: pd.DataFrame,
) -> pd.DataFrame:
    """
    Helper to aggregate features to bout-level for unsupervised analyses.

    :param pd.DataFrame data: DataFrame with features.
    :param List[str] clfs: Names of classifiers
    :param feature_names: Names of features
    :param str aggregator: Aggregation type, e.g., 'MEAN', 'MEDIAN'
    :param int min_bout_length: The length of the shortest allowed bout in milliseconds.
    :param pd.DataFrame video_info: Holding video fps, resolution etc.
    :return pd.DataFrame: Featurized data at aggregate bout level.
    """

    print("Calculating bout aggregate statistics...")

    def bout_aggregator_mp(frms, data, clf_name):
        bout_df = data.iloc[frms[0] : frms[1] + 1]
        bout_video, start_frm, end_frm = (
            bout_df["VIDEO"].values[0],
            bout_df["FRAME"].values[0],
            bout_df["FRAME"].values[-1],
        )
        if aggregator == "MEAN":
            agg_df = pd.DataFrame(bout_df[feature_names].mean()).T
            agg_df["PROBABILITY"] = bout_df[f"Probability_{clf_name}"].mean()
        elif aggregator == "MEDIAN":
            agg_df = pd.DataFrame(bout_df[feature_names].median()).T
            agg_df["PROBABILITY"] = bout_df[f"Probability_{clf_name}"].median()
        agg_df["CLASSIFIER"] = clf_name
        agg_df.insert(0, "END_FRAME", end_frm)
        agg_df.insert(0, "START_FRAME", start_frm)
        agg_df.insert(0, "VIDEO", bout_video)
        return agg_df

    output = []
    for cnt, video in enumerate(data["VIDEO"].unique()):
        print(
            f'Processing video {video}...({str(cnt+1)}/{str(len(data["VIDEO"].unique()))})'
        )
        video_df = data[data["VIDEO"] == video].reset_index(drop=True)
        for clf in clfs:
            _, _, fps = read_video_info(vid_info_df=video_info, video_name=video)
            bouts = detect_bouts(
                data_df=video_df, target_lst=[clf], fps=fps
            ).sort_values(by="Start_frame")
            bouts = bouts[bouts["Bout_time"] >= min_bout_length / 1000][
                ["Start_frame", "End_frame"]
            ].values
            if len(bouts) > 0:
                bouts = [x.tolist() for x in bouts]
                results = Parallel(n_jobs=-1, verbose=0, backend="loky")(
                    delayed(bout_aggregator_mp)(j, video_df, clf) for j in bouts
                )
                results = pd.concat(results, axis=0).sort_values(
                    by=["VIDEO", "START_FRAME"]
                )
                output.append(results)
    get_reusable_executor().shutdown(wait=True)

    return pd.concat(output, axis=0).reset_index(drop=True)



__author__ = "Simon Nilsson"

import itertools
import os
from typing import List

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_if_filepath_list_is_empty, check_int)
from simba.utils.enums import TagNames
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class TimeBinsMovementCalculator(ConfigReader, FeatureExtractionMixin):
    """
    Computes aggregate movement statistics in user-defined time-bins.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter int bin_length: Integer representing the time bin size in seconds.
    :parameter bool plots: If True, creates time-bin line plots representing the movement in each time-bin per video. Results are saved in the ``project_folder/logs/`` sub-directory.

    .. note::
        `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-4--analyze-machine-results>`__.

    .. image:: _static/img/TimeBinsMovementCalculator.png
       :width: 500
       :align: center

    Example
    ----------
    >>> timebin_movement_analyzer = TimeBinsMovementCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini', bin_length=60, plots=True, body_parts=['midpoint', 'mouth'])
    >>> timebin_movement_analyzer.run()
    """

    def __init__(
        self,
        config_path: str,
        bin_length: int,
        body_parts: List[str],
        plots: bool = False,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        self.bin_length, self.plots = bin_length, plots
        check_int(name="TIME BIN", value=bin_length, min_value=1)
        self.col_headers, self.bp_dict = [], {}
        for bp_cnt, bp in enumerate(body_parts):
            self.col_headers.extend((f"{bp}_x", f"{bp}_y"))
            animal_name = self.find_animal_name_from_body_part_name(
                bp_name=bp, bp_dict=self.animal_bp_dict
            )
            self.bp_dict[bp_cnt] = {animal_name: [f"{bp}_x", f"{bp}_y"]}
        check_if_filepath_list_is_empty(
            filepaths=self.outlier_corrected_paths,
            error_msg=f"SIMBA ERROR: Cannot analyze movement in time-bins, data directory {self.outlier_corrected_dir} is empty.",
        )
        self.animal_combinations = list(itertools.combinations(self.animal_bp_dict, 2))
        print(
            "Processing {} video(s)...".format(str(len(self.outlier_corrected_paths)))
        )

    def __create_plots(self):
        timer = SimbaTimer(start=True)
        print("Creating time-bin movement plots...")
        sns.set_style("whitegrid", {"grid.linestyle": "--"})
        plots_dir = os.path.join(
            self.project_path, "logs", f"time_bin_movement_plots_{self.datetime}"
        )
        if not os.path.exists(plots_dir):
            os.makedirs(plots_dir)
        for video_name in self.results["VIDEO"].unique():
            video_df = self.results.loc[
                (self.results["VIDEO"] == video_name)
                & (self.results["MEASUREMENT"] == "Movement (cm)")
            ]
            video_df["Time bin #"] = video_df["Time bin #"].astype(int)
            for body_part in video_df["BODY-PART"].unique():
                body_part_df = (
                    video_df[video_df["BODY-PART"] == body_part]
                    .reset_index(drop=True)
                    .sort_values(by=["Time bin #"])
                )
                body_part_df[f"Time bin # (bin length {self.bin_length}s)"] = (
                    body_part_df["Time bin #"]
                )
                line_plot = sns.lineplot(
                    data=body_part_df,
                    x=f"Time bin # (bin length {self.bin_length}s)",
                    y="VALUE",
                )
                plt.ylabel("Distance (cm)")
                self.plot_save_path = os.path.join(
                    plots_dir, f"{video_name}_{body_part}.png"
                )
                line_plot.figure.savefig(self.plot_save_path)
                plt.close()
        timer.stop_timer()
        stdout_success(
            msg=f"Time bin movement plots saved in {plots_dir}",
            elapsed_time=timer.elapsed_time_str,
            source=self.__class__.__name__,
        )

    def run(self):
        """
        Method for running the movement time-bin analysis. Results are stored in the ``project_folder/logs`` directory
        of the SimBA project.

        Returns
        ----------
        None
        """
        video_dict, out_df_lst = {}, []
        self.save_path = os.path.join(
            self.project_path,
            "logs",
            "Time_bins_movement_results_" + self.datetime + ".csv",
        )
        check_all_file_names_are_represented_in_video_log(
            video_info_df=self.video_info_df, data_paths=self.outlier_corrected_paths
        )
        for file_cnt, file_path in enumerate(self.outlier_corrected_paths):
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(file_path)
            print(
                f"Processing time-bin movements for video {video_name} ({str(file_cnt+1)}/{str(len(self.outlier_corrected_paths))})..."
            )
            video_dict[video_name] = {}
            video_settings, px_per_mm, fps = self.read_video_info(video_name=video_name)
            fps, self.movement_cols, self.velocity_cols = int(fps), set(), set()
            bin_length_frames = int(fps * self.bin_length)
            data_df = read_df(file_path, self.file_type, usecols=self.col_headers)
            data_df = self.create_shifted_df(df=data_df)
            for animal_data in self.bp_dict.values():
                name, bps = list(animal_data.keys())[0], list(animal_data.values())[0]
                bp_time_1, bp_time_2 = (
                    data_df[bps].values,
                    data_df[[f"{bps[0]}_shifted", f"{bps[1]}_shifted"]].values,
                )
                movement_data = pd.DataFrame(
                    self.framewise_euclidean_distance(
                        location_1=bp_time_1,
                        location_2=bp_time_2,
                        px_per_mm=px_per_mm,
                        centimeter=True,
                    ),
                    columns=["VALUE"],
                )
                results_df_lists = [
                    movement_data[i : i + bin_length_frames]
                    for i in range(0, movement_data.shape[0], bin_length_frames)
                ]
                indexed_df = []
                for bin, results in enumerate(results_df_lists):
                    time_bin_per_s = [
                        results[i : i + fps] for i in range(0, results.shape[0], fps)
                    ]
                    for second, df in enumerate(time_bin_per_s):
                        df["Time bin #"] = bin
                        df["Second"] = second
                        indexed_df.append(df)
                indexed_df = pd.concat(indexed_df, axis=0)
                velocity_df = (
                    indexed_df.groupby(["Time bin #", "Second"])["VALUE"]
                    .sum()
                    .reset_index()
                )
                velocity_df = (
                    velocity_df.groupby(["Time bin #"])["VALUE"].mean().reset_index()
                )
                velocity_df["ANIMAL"] = list(animal_data.keys())[0]
                velocity_df["BODY-PART"] = bps[0][:-2]
                velocity_df["MEASUREMENT"] = "Velocity (cm/s)"
                movement_df = (
                    indexed_df.groupby(["Time bin #"])["VALUE"].sum().reset_index()
                )
                movement_df["ANIMAL"] = list(animal_data.keys())[0]
                movement_df["BODY-PART"] = bps[0][:-2]
                movement_df["MEASUREMENT"] = "Movement (cm)"
                results = pd.concat([movement_df, velocity_df], axis=0)
                results["VIDEO"] = video_name
                out_df_lst.append(results)
            video_timer.stop_timer()
            print(
                f"Video {video_name} complete (elapsed time: {video_timer.elapsed_time_str}s)..."
            )
        self.results = pd.concat(out_df_lst, axis=0).sort_values(
            by=["VIDEO", "Time bin #", "MEASUREMENT", "ANIMAL"]
        )[["VIDEO", "Time bin #", "ANIMAL", "BODY-PART", "MEASUREMENT", "VALUE"]]
        self.results.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Movement time-bins results saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )
        if self.plots:
            self.__create_plots()


# test = TimeBinsMovementCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
#                                   bin_length=60, plots=True, body_parts=['midpoint', 'mouth'])
# test.run()

# test = TimeBinsMovementCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini',
#                                 bin_length=1, plots=True)
# test.analyze_movement()



__author__ = "Simon Nilsson"


# MODIFIED FROM https://gist.github.com/HaydenFaulkner/54318fd3e9b9bdb66c5440c44e4e08b8
# Medium article https://medium.com/@haydenfaulkner/extracting-frames-fast-from-a-video-using-opencv-and-python-73b9b7dc9661
# All cred to Hayden Faulkner, ta!

import multiprocessing
import os
from concurrent.futures import ProcessPoolExecutor

import cv2


def extract_frames(
    video_filename, frames_dir, overwrite=True, start=-1, end=-1, every=1
):
    capture = cv2.VideoCapture(video_filename)
    if start < 0:
        start = 0
    if end < 0:
        end = int(capture.get(cv2.CAP_PROP_FRAME_COUNT))

    capture.set(1, start)
    frame = start
    while_safety = 0
    saved_count = 0

    while frame < end:
        _, image = capture.read()

        if while_safety > 500:
            break
        if image is None:
            while_safety += 1
            continue  # skip

        if (
            frame % every == 0
        ):  # if this is a frame we want to write out based on the 'every' argument
            while_safety = 0  # reset the safety count
            save_path = os.path.join(
                frames_dir, "{:0d}.png".format(frame)
            )  # create the save path
            if (
                not os.path.exists(save_path) or overwrite
            ):  # if it doesn't exist or we want to overwrite anyways
                cv2.imwrite(save_path, image)  # save the extracted image
                saved_count += 1  # increment our counter by one

        frame += 1  # increment our frame count

    capture.release()  # after the while has finished close the capture

    return saved_count  # and return the count of the images we saved


def video_to_frames(
    video_filename, frames_dir, overwrite=True, every=1, chunk_size=1000
):
    capture = cv2.VideoCapture(video_filename)  # load the video
    total = int(capture.get(cv2.CAP_PROP_FRAME_COUNT))  # get its total frame count
    capture.release()  # release the capture straight away

    if total < 1:  # if video has no frames, might be and opencv error
        print("The video has no frames")
        return None  # return None

    frame_chunks = [
        [i, i + chunk_size] for i in range(0, total, chunk_size)
    ]  # split the frames into chunk lists
    frame_chunks[-1][-1] = min(
        frame_chunks[-1][-1], total
    )  # make sure last chunk has correct end frame

    print(
        "Extracting "
        + str(total)
        + " frames from "
        + str(os.path.basename(video_filename))
        + "..."
    )

    # execute across multiple cpu cores to speed up processing, get the count automatically
    with ProcessPoolExecutor(max_workers=multiprocessing.cpu_count()) as executor:
        futures = [
            executor.submit(
                extract_frames, video_filename, frames_dir, overwrite, f[0], f[1], every
            )
            for f in frame_chunks
        ]  # submit the processes: extract_frames(...)



import pickle

import networkx as nx
import numpy as np
import pyvista as pv
from fa2 import ForceAtlas2


class Graph3DPlotter(object):
    def __init__(self, networks_path: dict):
        with open(networks_path, "rb") as handle:
            self.networks = pickle.load(handle)

        force_atlas = ForceAtlas2(verbose=False)
        self.node_locations = {}
        for i in self.networks.keys():
            node_locations = force_atlas.forceatlas2_networkx_layout(
                self.networks[i], pos=None, iterations=100
            )
            for node in node_locations.keys():
                node_location = list(node_locations[node])
                node_location.append(0.0)
                node_locations[node] = node_location
            self.node_locations[i] = node_locations

    def plot(self, video_name: str):
        self.plotter = pv.Plotter()
        self.plotter.set_background("darkgrey", top="dodgerblue")
        node_locations = self.node_locations[video_name]
        node_array = np.full((len(node_locations.keys()), 3), 0.0)
        print(node_array)

        for cnt, animal_name in enumerate(node_locations):
            node_array[cnt] = np.array(node_locations[animal_name])
        node_cloud = pv.wrap(node_array)
        self.plotter.add_mesh(
            node_cloud,
            scalars=None,
            render_points_as_spheres=True,
            point_size=45,
            cmap="Accent",
            ambient=0.5,
            categories=True,
            name="non_target_baseline",
        )

        for edge in self.networks[video_name].edges.data():
            line_points = np.array(
                [
                    self.node_locations[video_name][edge[0]],
                    self.node_locations[video_name][edge[1]],
                ]
            )
            self.plotter.add_lines(line_points, color="yellow", width=edge[2]["weight"])

        self.plotter.show()


test = Graph3DPlotter(networks_path="/Users/simon/Desktop/envs/simba_dev/Test.pickle")
test.plot(video_name="termites_test")



__author__ = "Simon Nilsson"

import math
import os
from collections import defaultdict
from copy import deepcopy
from typing import Union

import numpy as np
import pandas as pd
import scipy
from scipy.spatial import ConvexHull

from simba.feature_extractors.perimeter_jit import jitted_hull
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


class ExtractFeaturesFrom7bps(ConfigReader, FeatureExtractionMixin):
    """
    Extracts hard-coded set of features from pose-estimation data from single animals with 7 tracked body-parts.
    Results are stored in the `project_folder/csv/features_extracted` directory of the SimBA project.

    :parameter str config_path: path to SimBA project config file in Configparser format

    .. note::
      `Feature extraction tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-5-extract-features>`_.
      `Expected pose configuration <https://github.com/sgoldenlab/simba/blob/master/simba/pose_configurations/schematics/2.png>`_
       .. image:: _static/img/pose_configurations/2.png
          :width: 150
          :align: center

    Examples
    ----------
    >>> feature_extractor = ExtractFeaturesFrom7bps(config_path='MyProjectConfig')
    >>> feature_extractor.run()

    """

    def __init__(self, config_path: Union[str, os.PathLike]):
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        self.in_headers = self.get_feature_extraction_headers(
            pose="1 animal 7 body-parts"
        )
        self.mouse_p_headers = [x for x in self.in_headers if x[-2:] == "_p"]
        self.mouse_headers = [x for x in self.in_headers if x[-2:] != "_p"]
        print(
            "Extracting features from {} file(s)...".format(str(len(self.files_found)))
        )

    def run(self):
        """
        Method to compute and save features to disk. Results are saved in the `project_folder/csv/features_extracted`
        directory of the SimBA project.

        Returns
        -------
        None
        """
        for file_cnt, file_path in enumerate(self.files_found):
            _, self.video_name, _ = get_fn_ext(file_path)
            video_settings, self.px_per_mm, fps = self.read_video_info(
                video_name=self.video_name
            )
            roll_windows = []
            for window in self.roll_windows_values:
                roll_windows.append(int(fps / window))
            self.in_data = (
                read_df(file_path, self.file_type)
                .fillna(0)
                .apply(pd.to_numeric)
                .reset_index(drop=True)
            )
            print(
                "Processing {} ({} frames)...".format(
                    self.video_name, str(len(self.in_data))
                )
            )
            self.in_data = self.insert_default_headers_for_feature_extraction(
                df=self.in_data,
                headers=self.in_headers,
                pose_config="7 body-parts",
                filename=file_path,
            )
            self.out_data = deepcopy(self.in_data)
            mouse_array = np.reshape(
                self.out_data[self.mouse_headers].values,
                (len(self.out_data / 2), -1, 2),
            ).astype(np.float32)
            self.out_data["Mouse_poly_area"] = (
                jitted_hull(points=mouse_array, target="perimeter") / self.px_per_mm
            )
            self.in_data = self.create_shifted_df(df=self.out_data)
            self.out_data["Mouse_nose_to_tail"] = self.euclidean_distance(
                self.in_data["Nose_x"].values,
                self.in_data["Tail_base_x"].values,
                self.in_data["Nose_y"].values,
                self.in_data["Tail_base_y"].values,
                self.px_per_mm,
            )
            self.out_data["Mouse_width"] = self.euclidean_distance(
                self.in_data["Lat_left_x"].values,
                self.in_data["Lat_right_x"].values,
                self.in_data["Lat_left_y"].values,
                self.in_data["Lat_right_y"].values,
                self.px_per_mm,
            )
            self.out_data["Mouse_Ear_distance"] = self.euclidean_distance(
                self.in_data["Ear_left_x"].values,
                self.in_data["Ear_right_x"].values,
                self.in_data["Ear_left_y"].values,
                self.in_data["Ear_right_y"].values,
                self.px_per_mm,
            )
            self.out_data["Mouse_Nose_to_centroid"] = self.euclidean_distance(
                self.in_data["Nose_x"].values,
                self.in_data["Center_x"].values,
                self.in_data["Nose_y"].values,
                self.in_data["Center_y"].values,
                self.px_per_mm,
            )
            self.out_data["Mouse_Nose_to_lateral_left"] = self.euclidean_distance(
                self.in_data["Nose_x"].values,
                self.in_data["Lat_left_x"].values,
                self.in_data["Nose_y"].values,
                self.in_data["Lat_left_y"].values,
                self.px_per_mm,
            )
            self.out_data["Mouse_Nose_to_lateral_right"] = self.euclidean_distance(
                self.in_data["Nose_x"].values,
                self.in_data["Lat_right_x"].values,
                self.in_data["Nose_y"].values,
                self.in_data["Lat_right_y"].values,
                self.px_per_mm,
            )
            self.out_data["Mouse_Centroid_to_lateral_left"] = self.euclidean_distance(
                self.in_data["Center_x"].values,
                self.in_data["Lat_left_x"].values,
                self.in_data["Center_y"].values,
                self.in_data["Lat_left_y"].values,
                self.px_per_mm,
            )
            self.out_data["Mouse_Centroid_to_lateral_right"] = self.euclidean_distance(
                self.in_data["Center_x"].values,
                self.in_data["Lat_right_x"].values,
                self.in_data["Center_y"].values,
                self.in_data["Lat_right_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_mouse_centroid"] = self.euclidean_distance(
                self.in_data["Center_x_shifted"].values,
                self.in_data["Center_x"].values,
                self.in_data["Center_y_shifted"].values,
                self.in_data["Center_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_mouse_nose"] = self.euclidean_distance(
                self.in_data["Nose_x_shifted"].values,
                self.in_data["Nose_x"].values,
                self.in_data["Nose_y_shifted"].values,
                self.in_data["Nose_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_mouse_tail_base"] = self.euclidean_distance(
                self.in_data["Tail_base_x_shifted"].values,
                self.in_data["Tail_base_x"].values,
                self.in_data["Tail_base_y_shifted"].values,
                self.in_data["Tail_base_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_mouse_left_ear"] = self.euclidean_distance(
                self.in_data["Ear_left_x_shifted"].values,
                self.in_data["Ear_left_x"].values,
                self.in_data["Ear_left_y_shifted"].values,
                self.in_data["Ear_left_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_mouse_right_ear"] = self.euclidean_distance(
                self.in_data["Ear_right_x_shifted"].values,
                self.in_data["Ear_right_x"].values,
                self.in_data["Ear_right_y_shifted"].values,
                self.in_data["Ear_right_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_mouse_lateral_left"] = self.euclidean_distance(
                self.in_data["Lat_left_x_shifted"].values,
                self.in_data["Lat_left_x"].values,
                self.in_data["Lat_left_y_shifted"].values,
                self.in_data["Lat_left_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_mouse_lateral_right"] = self.euclidean_distance(
                self.in_data["Lat_right_x_shifted"].values,
                self.in_data["Lat_right_x"].values,
                self.in_data["Lat_right_y_shifted"].values,
                self.in_data["Lat_right_y"].values,
                self.px_per_mm,
            )

            self.out_data["Mouse_polygon_size_change"] = (
                self.in_data["Mouse_poly_area_shifted"]
                - self.in_data["Mouse_poly_area"]
            )

            print("Calculating hull variables...")
            mouse_array = self.in_data[self.mouse_headers].to_numpy()
            self.hull_dict = defaultdict(list)
            for cnt, animal_frm in enumerate(mouse_array):
                animal_frm = np.reshape(animal_frm, (-1, 2))
                animal_dists = scipy.spatial.distance.cdist(
                    animal_frm, animal_frm, metric="euclidean"
                )
                animal_dists = animal_dists[animal_dists != 0]
                self.hull_dict["M1_largest_euclidean_distance_hull"].append(
                    np.amax(animal_dists, initial=0) / self.px_per_mm
                )
                self.hull_dict["M1_smallest_euclidean_distance_hull"].append(
                    np.amin(
                        animal_dists,
                        initial=self.hull_dict["M1_largest_euclidean_distance_hull"][
                            -1
                        ],
                    )
                    / self.px_per_mm
                )
                self.hull_dict["M1_mean_euclidean_distance_hull"].append(
                    np.mean(animal_dists) / self.px_per_mm
                )
                self.hull_dict["M1_sum_euclidean_distance_hull"].append(
                    np.sum(animal_dists) / self.px_per_mm
                )
            for k, v in self.hull_dict.items():
                self.out_data[k] = v

            self.out_data["Total_movement_all_bodyparts_M1"] = self.out_data[
                [
                    "Movement_mouse_centroid",
                    "Movement_mouse_nose",
                    "Movement_mouse_left_ear",
                    "Movement_mouse_right_ear",
                    "Movement_mouse_lateral_left",
                    "Movement_mouse_lateral_right",
                ]
            ].sum(axis=1)

            print("Calculating rolling windows: medians, medians, and sums...")
            for window in self.roll_windows_values:
                col_name = "Mouse1_width_median_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Mouse_width"]
                    .rolling(int(window), min_periods=1)
                    .median()
                )
                col_name = "Mouse1_width_mean_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Mouse_width"]
                    .rolling(int(window), min_periods=1)
                    .mean()
                )
                col_name = "Mouse1_width_sum_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Mouse_width"]
                    .rolling(int(window), min_periods=1)
                    .sum()
                )

            for window in self.roll_windows_values:
                col_name = "Mouse1_mean_euclid_distances_median_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["M1_mean_euclidean_distance_hull"]
                    .rolling(int(window), min_periods=1)
                    .median()
                )
                col_name = "Mouse1_mean_euclid_distances_mean_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["M1_mean_euclidean_distance_hull"]
                    .rolling(int(window), min_periods=1)
                    .mean()
                )
                col_name = "Mouse1_mean_euclid_distances_sum_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["M1_mean_euclidean_distance_hull"]
                    .rolling(int(window), min_periods=1)
                    .sum()
                )

            for window in self.roll_windows_values:
                col_name = "Mouse1_smallest_euclid_distances_median_{}".format(
                    str(window)
                )
                self.out_data[col_name] = (
                    self.out_data["M1_smallest_euclidean_distance_hull"]
                    .rolling(int(window), min_periods=1)
                    .median()
                )
                col_name = "Mouse1_smallest_euclid_distances_mean_{}".format(
                    str(window)
                )
                self.out_data[col_name] = (
                    self.out_data["M1_smallest_euclidean_distance_hull"]
                    .rolling(int(window), min_periods=1)
                    .mean()
                )
                col_name = "Mouse1_smallest_euclid_distances_sum_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["M1_smallest_euclidean_distance_hull"]
                    .rolling(int(window), min_periods=1)
                    .sum()
                )

            for window in self.roll_windows_values:
                col_name = "Mouse1_largest_euclid_distances_median_{}".format(
                    str(window)
                )
                self.out_data[col_name] = (
                    self.out_data["M1_largest_euclidean_distance_hull"]
                    .rolling(int(window), min_periods=1)
                    .median()
                )
                col_name = "Mouse1_largest_euclid_distances_mean_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["M1_largest_euclidean_distance_hull"]
                    .rolling(int(window), min_periods=1)
                    .mean()
                )
                col_name = "Mouse1_largest_euclid_distances_sum_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["M1_largest_euclidean_distance_hull"]
                    .rolling(int(window), min_periods=1)
                    .sum()
                )

            for window in self.roll_windows_values:
                col_name = "Tail_base_movement_M1_median_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Movement_mouse_tail_base"]
                    .rolling(int(window), min_periods=1)
                    .median()
                )
                col_name = "Tail_base_movement_M1_mean_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Movement_mouse_tail_base"]
                    .rolling(int(window), min_periods=1)
                    .mean()
                )
                col_name = "Tail_base_movement_M1_sum_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Movement_mouse_tail_base"]
                    .rolling(int(window), min_periods=1)
                    .sum()
                )

            for window in self.roll_windows_values:
                col_name = "Centroid_movement_M1_median_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Movement_mouse_centroid"]
                    .rolling(int(window), min_periods=1)
                    .median()
                )
                col_name = "Centroid_movement_M1_mean_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Movement_mouse_centroid"]
                    .rolling(int(window), min_periods=1)
                    .mean()
                )
                col_name = "Centroid_movement_M1_sum_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Movement_mouse_centroid"]
                    .rolling(int(window), min_periods=1)
                    .sum()
                )

            for window in self.roll_windows_values:
                col_name = "Nose_movement_M1_median_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Movement_mouse_nose"]
                    .rolling(int(window), min_periods=1)
                    .median()
                )
                col_name = "Nose_movement_M1_mean_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Movement_mouse_nose"]
                    .rolling(int(window), min_periods=1)
                    .mean()
                )
                col_name = "Nose_movement_M1_sum_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Movement_mouse_nose"]
                    .rolling(int(window), min_periods=1)
                    .sum()
                )

            for window in self.roll_windows_values:
                col_name = "Total_movement_M1_median_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Total_movement_all_bodyparts_M1"]
                    .rolling(int(window), min_periods=1)
                    .median()
                )
                col_name = "Total_movement_M1_mean_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Total_movement_all_bodyparts_M1"]
                    .rolling(int(window), min_periods=1)
                    .mean()
                )
                col_name = "Total_movement_M1_sum_{}".format(str(window))
                self.out_data[col_name] = (
                    self.out_data["Total_movement_all_bodyparts_M1"]
                    .rolling(int(window), min_periods=1)
                    .sum()
                )

            print("Calculating angles...")
            self.out_data["Mouse_angle"] = self.angle3pt_serialized(
                data=self.out_data[
                    [
                        "Nose_x",
                        "Nose_y",
                        "Center_x",
                        "Center_y",
                        "Tail_base_x",
                        "Tail_base_y",
                    ]
                ].values
            )

            print("Calculating deviations...")
            self.out_data["Total_movement_all_bodyparts_deviation"] = (
                self.out_data["Total_movement_all_bodyparts_M1"].mean()
                - self.out_data["Total_movement_all_bodyparts_M1"]
            )
            self.out_data["M1_smallest_euclid_distances_hull_deviation"] = (
                self.out_data["M1_smallest_euclidean_distance_hull"].mean()
                - self.out_data["M1_smallest_euclidean_distance_hull"]
            )
            self.out_data["M1_largest_euclid_distances_hull_deviation"] = (
                self.out_data["M1_largest_euclidean_distance_hull"].mean()
                - self.out_data["M1_largest_euclidean_distance_hull"]
            )
            self.out_data["M1_mean_euclid_distances_hull_deviation"] = (
                self.out_data["M1_mean_euclidean_distance_hull"].mean()
                - self.out_data["M1_mean_euclidean_distance_hull"]
            )
            self.out_data["Total_angle_deviation"] = (
                self.out_data["Mouse_angle"].mean() - self.out_data["Mouse_angle"]
            )
            self.out_data["Movement_mouse_deviation_centroid"] = (
                self.out_data["Movement_mouse_centroid"].mean()
                - self.out_data["Movement_mouse_centroid"]
            )
            self.out_data["Mouse_polygon_deviation"] = (
                self.out_data["Mouse_poly_area"].mean()
                - self.out_data["Mouse_poly_area"]
            )

            for window in self.roll_windows_values:
                col_name = "Mouse1_smallest_euclid_distances_mean_{}".format(
                    str(window)
                )
                deviation_col_name = col_name + "_deviation"
                self.out_data[deviation_col_name] = (
                    self.out_data[col_name].mean() - self.out_data[col_name]
                )

            for window in self.roll_windows_values:
                col_name = "Mouse1_largest_euclid_distances_mean_{}".format(str(window))
                deviation_col_name = col_name + "_deviation"
                self.out_data[deviation_col_name] = (
                    self.out_data[col_name].mean() - self.out_data[col_name]
                )

            for window in self.roll_windows_values:
                col_name = "Mouse1_mean_euclid_distances_mean_{}".format(str(window))
                deviation_col_name = col_name + "_deviation"
                self.out_data[deviation_col_name] = (
                    self.out_data[col_name].mean() - self.out_data[col_name]
                )

            self.out_data["Movement_mouse_percentile_rank"] = self.out_data[
                "Movement_mouse_centroid"
            ].rank(pct=True)
            self.out_data["Movement_mouse_deviation_percentile_rank"] = self.out_data[
                "Movement_mouse_deviation_centroid"
            ].rank(pct=True)

            for window in self.roll_windows_values:
                col_name = "Total_movement_M1_mean_{}".format(str(window))
                deviation_col_name = col_name + "_percentile_rank"
                self.out_data[deviation_col_name] = (
                    self.out_data[col_name].mean() - self.out_data[col_name]
                )

            for window in self.roll_windows_values:
                col_name = "Mouse1_mean_euclid_distances_mean_{}".format(str(window))
                deviation_col_name = col_name + "_percentile_rank"
                self.out_data[deviation_col_name] = (
                    self.out_data[col_name].mean() - self.out_data[col_name]
                )

            for window in self.roll_windows_values:
                col_name = "Mouse1_smallest_euclid_distances_mean_{}".format(
                    str(window)
                )
                deviation_col_name = col_name + "_percentile_rank"
                self.out_data[deviation_col_name] = (
                    self.out_data[col_name].mean() - self.out_data[col_name]
                )

            for window in self.roll_windows_values:
                col_name = "Mouse1_largest_euclid_distances_mean_{}".format(str(window))
                deviation_col_name = col_name + "_percentile_rank"
                self.out_data[deviation_col_name] = (
                    self.out_data[col_name].mean() - self.out_data[col_name]
                )

            print("Calculating path tortuosities...")
            as_strided = np.lib.stride_tricks.as_strided
            win_size = 3
            centroid_lst_mouse_x = as_strided(
                self.out_data["Center_x"],
                (len(self.out_data) - (win_size - 1), win_size),
                (self.out_data["Center_x"].values.strides * 2),
            )
            centroid_lst_mouse_y = as_strided(
                self.out_data["Center_y"],
                (len(self.out_data) - (win_size - 1), win_size),
                (self.out_data["Center_y"].values.strides * 2),
            )

            for window in self.roll_windows_values:
                start, end = 0, 0 + int(window)
                tortuosities_results = defaultdict(list)
                for frame in range(len(self.out_data)):
                    tortuosities_dict = defaultdict(list)
                    c_centroid_lst_mouse_x, c_centroid_lst_mouse_y = (
                        centroid_lst_mouse_x[start:end],
                        centroid_lst_mouse_y[start:end],
                    )
                    for frame_in_window in range(len(c_centroid_lst_mouse_x)):
                        move_angle_mouse_ = self.angle3pt(
                            c_centroid_lst_mouse_x[frame_in_window][0],
                            c_centroid_lst_mouse_y[frame_in_window][0],
                            c_centroid_lst_mouse_x[frame_in_window][1],
                            c_centroid_lst_mouse_y[frame_in_window][1],
                            c_centroid_lst_mouse_x[frame_in_window][2],
                            c_centroid_lst_mouse_y[frame_in_window][2],
                        )
                        tortuosities_dict["Animal_1"].append(move_angle_mouse_)
                    tortuosities_results["Animal_1"].append(
                        sum(tortuosities_dict["Animal_1"]) / (2 * math.pi)
                    )
                    start += 1
                    end += 1
                col_name = "Tortuosity_Mouse1_{}".format(str(window))
                self.out_data[col_name] = tortuosities_results["Animal_1"]

            print("Calculating pose probability scores...")
            self.out_data["Sum_probabilities"] = (
                self.out_data["Ear_left_p"]
                + self.out_data["Ear_right_p"]
                + self.out_data["Nose_p"]
                + self.out_data["Tail_base_p"]
            )
            self.out_data["Sum_probabilities_deviation"] = (
                self.out_data["Sum_probabilities"].mean()
                - self.out_data["Sum_probabilities"]
            )
            self.out_data["Sum_probabilities_deviation_percentile_rank"] = (
                self.out_data["Sum_probabilities_deviation"].rank(pct=True)
            )
            self.out_data["Sum_probabilities_percentile_rank"] = self.out_data[
                "Sum_probabilities_deviation_percentile_rank"
            ].rank(pct=True)
            results = pd.DataFrame(
                self.count_values_in_range(
                    data=self.out_data.filter(self.mouse_p_headers).values,
                    ranges=np.array([[0.0, 0.1], [0.0, 0.5], [0.0, 0.75]]),
                ),
                columns=[
                    "Low_prob_detections_0.1",
                    "Low_prob_detections_0.5",
                    "Low_prob_detections_0.75",
                ],
            )
            self.out_data = pd.concat([self.out_data, results], axis=1)

            self.out_data = self.out_data.reset_index(drop=True).fillna(0)
            save_path = os.path.join(
                self.save_dir, self.video_name + "." + self.file_type
            )
            write_df(df=self.out_data, file_type=self.file_type, save_path=save_path)
            print(
                f"Feature extraction complete for {self.video_name} ({file_cnt + 1}/{len(self.files_found)})..."
            )

        self.timer.stop_timer()
        stdout_success(
            msg="All features extracted. Results are stored in the project_folder/csv/features_extracted directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = ExtractFeaturesFrom7bps(config_path='/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/project_config.ini')
# test.run()



import itertools
import pickle

import networkx as nx
import pandas as pd
from pyvis.network import Network

from simba.misc_tools import create_single_color_lst


class GraphCreator(object):
    def __init__(
        self,
        animals: list,
        data_path: str,
        edge_weight_settings: dict,
        filter_settings: dict,
    ):
        self.data = pd.read_csv(data_path)
        self.edge_weight_settings, self.animal_names = edge_weight_settings, animals
        self.filter_settings = filter_settings

    def __weigh_time(self, df: pd.DataFrame):
        results = pd.DataFrame(columns=["source", "target", "value"])
        if self.edge_weight_settings["Aggregation_method"] is "Sum":
            results.loc[len(results)] = [
                df["ROI 1"].values[0],
                df["ROI 2"].values[0],
                df["BOUT TIME (s)"].sum(),
            ]
        elif self.edge_weight_settings["Aggregation_method"] is "Mean":
            results.loc[len(results)] = [
                df["ROI 1"].values[0],
                df["ROI 2"].values[0],
                df["BOUT TIME (s)"].mean(),
            ]
        return results

    def __weigh_events(self, df: pd.DataFrame):
        results = pd.DataFrame(columns=["source", "target", "value"])
        results.loc[len(results)] = [
            df["ROI 1"].values[0],
            df["ROI 2"].values[0],
            len(df),
        ]

        return results

    def save(self, save_path: str):
        if not hasattr(self, "graphs"):
            raise ValueError()
        with open(save_path, "wb") as f:
            pickle.dump(self.graphs, f, protocol=pickle.HIGHEST_PROTOCOL)

    def run(self):
        self.graphs = {}
        for video_name in self.data["VIDEO"].unique():
            G = nx.Graph()
            for animal in self.animal_names:
                G.add_node(animal)
            weight_df_lst = []
            node_pairs = list(itertools.combinations(self.animal_names, 2))
            for node_pair in node_pairs:
                node_pair_df = self.data.loc[
                    (self.data["ROI 1"].isin(node_pair))
                    & (self.data["ROI 2"].isin(node_pair))
                ]

                if self.edge_weight_settings["Variable"] == "Time":
                    if self.filter_settings["Time_threshold"] is not None:
                        node_pair_df = node_pair_df[
                            node_pair_df["BOUT TIME (s)"]
                            > self.filter_settings["Time_threshold"]
                        ]
                    if len(node_pair_df) > 0:
                        weight_df_lst.append(self.__weigh_time(df=node_pair_df))

                if self.edge_weight_settings["Variable"] == "Counts":
                    if len(node_pair_df) > 0:
                        weight_df_lst.append(self.__weigh_events(df=node_pair_df))

            weight_arr = pd.concat(weight_df_lst, axis=0).reset_index(drop=True).values
            for i in range(weight_arr.shape[0]):
                G.add_edge(weight_arr[i][0], weight_arr[i][1], weight=weight_arr[i][2])
            self.graphs[video_name] = G

    def page_rank(self):
        for video_name, G in self.graphs.items():
            return nx.pagerank(G, alpha=0.9, max_iter=600, weight="weight")

    def visualize(
        self,
        node_colors: dict or None,
        node_size: dict or str or None,
        style_attr: dict = {"size": ["500px", "500px"]},
    ):
        for video_name, G in self.graphs.items():
            if node_colors is not None:
                clrs = create_single_color_lst(
                    pallete_name="magma", increments=len(self.animal_names), as_hex=True
                )
                node_colors = {
                    k: v
                    for k, v in sorted(
                        node_colors.items(), key=lambda item: item[1], reverse=True
                    )
                }
                for node_cnt, node_name in enumerate(node_colors.keys()):
                    node_colors[node_name] = clrs[node_cnt]

            network_graph = Network(style_attr["size"][0], style_attr["size"][1])
            for node_name, node_attrs in G.nodes(data=True):
                network_graph.add_node(node_name, color=node_colors[node_name])

            for source, target, edge_attrs in G.edges(data=True):
                edge_attrs["value"] = edge_attrs["weight"]
                network_graph.add_edge(str(source), str(target), **edge_attrs)

            network_graph.save_graph("nx.html")


test = GraphCreator(
    data_path="/Users/simon/Desktop/envs/simba_dev/tests/test_data/misc_test_files/termite_rois.csv",
    animals=["Animal_1", "Animal_2", "Animal_3", "Animal_4", "Animal_5"],
    edge_weight_settings={"Variable": "Time", "Aggregation_method": "Sum"},
    filter_settings={"Time_threshold": 1.0, "Count_threshold": None},
)

test.run()
test.save(save_path="Test.pickle")


#
# pr = test.page_rank()
# test.visualize(node_colors=pr, node_size=pr)



__author__ = "Simon Nilsson"

import os

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


class OutlierCorrectionSkipper(ConfigReader):
    """
    Skip outlier correction in SimBA projects.

    :parameter str config_path: path to SimBA project config file in Configparser format

    Examples
    ----------
    >>> _ = OutlierCorrectionSkipper(config_path='MyProjectConfig').run()
    """

    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if not os.path.exists(self.outlier_corrected_dir):
            os.makedirs(self.outlier_corrected_dir)
        check_if_filepath_list_is_empty(
            filepaths=self.input_csv_paths,
            error_msg=f"No files found in {self.input_csv_dir}.",
        )
        print(f"Processing {len(self.input_csv_paths)} file(s)...")

    def run(self):
        """
        Standardizes pose-estimation data (i.e., headers) from different pose-estimation packages.
        Results are stored in the project_folder/csv/outlier_corrected_movement_location` directory of
        the SimBA project
        """

        for file_cnt, file_path in enumerate(self.input_csv_paths):
            video_timer = SimbaTimer(start=True)
            _, video_name, ext = get_fn_ext(file_path)
            data_df = read_df(
                file_path=file_path, file_type=self.file_type, check_multiindex=True
            )
            if "scorer" in data_df.columns:
                data_df = data_df.set_index("scorer")

            data_df = self.insert_column_headers_for_outlier_correction(
                data_df=data_df, new_headers=self.bp_headers, filepath=file_path
            )
            data_df.index.name = None
            save_path = os.path.join(
                self.outlier_corrected_dir, video_name + "." + self.file_type
            )
            write_df(df=data_df, file_type=self.file_type, save_path=save_path)
            video_timer.stop_timer()
            print(
                f"Skipped outlier correction for video {video_name} (elapsed time {video_timer.elapsed_time_str}s)..."
            )
        self.timer.stop_timer()
        stdout_success(
            msg=f"Skipped outlier correction for {len(self.input_csv_paths)} files",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = OutlierCorrectionSkipper(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini')
# test.run()



import datetime
import glob
import json
import os
import pathlib
import shutil
import subprocess

import cv2

import simba
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.read_write import get_video_meta_data


class FFMPEGCommandCreator(object):
    """
    Execute FFmpeg commands from instructions stored in json format.

    Parameters
    ----------
    json_path: str
        path to json file storing FFmpeg instructions as created by ``simba.batch_process_vides.BatchProcessFrame``.

    Notes
    ----------
    `Batch pre-process tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial_process_videos.md>`__.
    `Example expected JSON file <https://github.com/sgoldenlab/simba/blob/master/misc/batch_process_log.json>`__.

    Examples
    ----------
    >>> ffmpeg_executor = FFMPEGCommandCreator(json_path='MyJsonFilePath')
    >>> ffmpeg_executor.crop_videos()
    >>> ffmpeg_executor.clip_videos()
    >>> ffmpeg_executor.downsample_videos()
    >>> ffmpeg_executor.apply_fps()
    >>> ffmpeg_executor.apply_grayscale()
    >>> ffmpeg_executor.apply_frame_count()
    >>> ffmpeg_executor.apply_clahe()
    >>> ffmpeg_executor.move_all_processed_files_to_output_folder()
    """

    def __init__(self, json_path: str):
        check_file_exist_and_readable(json_path)
        with open(json_path, "r") as fp:
            self.video_dict = json.load(fp)
        self.input_dir = self.video_dict["meta_data"]["in_dir"]
        self.out_dir = self.video_dict["meta_data"]["out_dir"]
        if "gpu" in list(self.video_dict["meta_data"].keys()):
            self.gpu = self.video_dict["meta_data"]["gpu"]
            self.quality = "medium"
        else:
            self.gpu = False
            self.quality = 23
        self.temp_dir = os.path.join(self.out_dir, "temp")
        self.time_format = "%H:%M:%S"
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)
        os.makedirs(self.temp_dir)
        self.copy_videos_to_temp_dir()

    def replace_files_in_temp(self):
        processed_files = glob.glob(self.process_dir + "/*")
        for file_path in processed_files:
            file_path_basename = os.path.basename(file_path)
            shutil.move(file_path, os.path.join(self.temp_dir, file_path_basename))

    def create_process_dir(self):
        self.process_dir = os.path.join(self.temp_dir, "process_dir")
        if os.path.exists(self.process_dir):
            shutil.rmtree(self.process_dir)
        os.makedirs(self.process_dir)

    def find_relevant_videos(self, variable=None):
        videos = {}
        for video, video_data in self.video_dict["video_data"].items():
            if video_data[variable]:
                video_basename = (
                    video_data["video_info"]["video_name"]
                    + video_data["video_info"]["extension"]
                )
                videos[video_data["video_info"]["video_name"]] = {}
                videos[video_data["video_info"]["video_name"]]["path"] = os.path.join(
                    self.temp_dir, video_basename
                )
                videos[video_data["video_info"]["video_name"]]["settings"] = video_data[
                    variable + "_settings"
                ]
                videos[video_data["video_info"]["video_name"]]["last_operation"] = (
                    video_data["last_operation"]
                )
                videos[video_data["video_info"]["video_name"]]["output_quality"] = (
                    video_data["output_quality"]
                )
        return videos

    def downsample_videos(self):
        self.videos_to_downsample = self.find_relevant_videos(variable="downsample")
        self.create_process_dir()
        for video, video_data in self.videos_to_downsample.items():
            print(f"Down-sampling {video}...")
            if video_data["last_operation"] == "downsample":
                self.quality = video_data["output_quality"]
            in_path, out_path = video_data["path"], os.path.join(
                self.process_dir, os.path.basename(video_data["path"])
            )
            width, height = str(video_data["settings"]["width"]), str(
                video_data["settings"]["height"]
            )
            if self.gpu:
                command = f'ffmpeg -y -hwaccel auto -c:v h264_cuvid -i "{in_path}" -vf "scale=w={width}:h={height}" -c:v h264_nvenc -preset {self.quality} -hide_banner -loglevel error "{out_path}"'
            else:
                command = f'ffmpeg -i "{in_path}" -vf scale={width}:{height} "{out_path}" -crf {self.quality} -hide_banner -loglevel error'
            subprocess.call(command, shell=True, stdout=subprocess.PIPE)
        self.replace_files_in_temp()
        print("Downsampling complete...")

    def clip_videos(self):
        self.videos_to_clip = self.find_relevant_videos(variable="clip")
        self.create_process_dir()
        for video, video_data in self.videos_to_clip.items():
            print("Clipping {}...".format(video))
            if video_data["last_operation"] == "clip":
                self.quality = video_data["output_quality"]
            in_path, out_path = video_data["path"], os.path.join(
                self.process_dir, os.path.basename(video_data["path"])
            )
            start_time, end_time = str(video_data["settings"]["start"]).replace(
                " ", ""
            ), str(video_data["settings"]["stop"]).replace(" ", "")
            start_time_shift, end_time = datetime.datetime.strptime(
                start_time, self.time_format
            ), datetime.datetime.strptime(end_time, self.time_format)
            time_difference = str(end_time - start_time_shift)
            if self.gpu:
                command = f'ffmpeg -hwaccel auto -i "{in_path}" -ss {start_time} -t {time_difference} -c:v h264_nvenc -async 1 "{out_path}" -preset {self.quality} -hide_banner -loglevel error'
            else:
                command = f'ffmpeg -i "{in_path}" -ss {start_time} -t {time_difference} -async 1 -crf {self.quality} -c:a copy "{out_path}" -hide_banner -loglevel error'
            subprocess.call(command, shell=True, stdout=subprocess.PIPE)
        self.replace_files_in_temp()
        print("Clipping complete...")

    def apply_fps(self):
        self.videos_to_change_fps = self.find_relevant_videos(variable="fps")
        self.create_process_dir()
        for video, video_data in self.videos_to_change_fps.items():
            print("Changing FPS {}...".format(video))
            if video_data["last_operation"] == "fps":
                self.quality = video_data["output_quality"]
            in_path, out_path = video_data["path"], os.path.join(
                self.process_dir, os.path.basename(video_data["path"])
            )
            fps = str(video_data["settings"]["fps"])
            if self.gpu:
                command = f'ffmpeg -hwaccel auto -c:v h264_cuvid -i "{in_path}" -vf "fps={fps}" -c:v h264_nvenc -preset {self.quality} -c:a copy "{out_path}" -hide_banner -loglevel error'
            else:
                command = f'ffmpeg -i "{in_path}" -crf {self.quality} -filter:v fps=fps={fps} "{out_path}"'
            subprocess.call(command, shell=True, stdout=subprocess.PIPE)
        self.replace_files_in_temp()
        print("FPS conversion complete...")

    def apply_grayscale(self):
        self.videos_to_greyscale = self.find_relevant_videos(variable="grayscale")
        self.create_process_dir()
        for video, video_data in self.videos_to_greyscale.items():
            print(f"Applying grayscale {video}...")
            if video_data["last_operation"] == "grayscale":
                self.quality = video_data["output_quality"]
            in_path, out_path = video_data["path"], os.path.join(
                self.process_dir, os.path.basename(video_data["path"])
            )
            if self.gpu:
                command = f'ffmpeg -hwaccel auto -c:v h264_cuvid -i "{in_path}" -vf "hwupload_cuda,hwdownload,format=nv12,format=gray" -c:v h264_nvenc -preset {self.quality} -c:a copy "{out_path}" -hide_banner -loglevel error'
            else:
                command = f'ffmpeg -i "{in_path}" -crf 17 -vf hue=s=0 "{out_path}" -hide_banner -loglevel error'
            subprocess.call(command, shell=True, stdout=subprocess.PIPE)
        self.replace_files_in_temp()
        print("Grayscale complete...")

    def apply_frame_count(self):
        self.videos_to_frm_cnt = self.find_relevant_videos(variable="frame_cnt")
        self.create_process_dir()
        for video, video_data in self.videos_to_frm_cnt.items():
            print(f"Applying frame count print {video}...")
            if video_data["last_operation"] == "frame_cnt":
                self.quality = video_data["output_quality"]
            in_path, out_path = video_data["path"], os.path.join(
                self.process_dir, os.path.basename(video_data["path"])
            )
            try:
                if self.gpu:
                    command = f'ffmpeg -hwaccel auto -c:v h264_cuvid -i "{in_path}" -vf "drawtext==fontfile=Arial.ttf:text=%{{n}}:x=(w-tw)/2:y=h-th-10:fontcolor=white:box=1:boxcolor=white@0.5" -c:v h264_nvenc -preset {self.quality} -c:a copy "{out_path}" -y -hide_banner -loglevel error'
                else:
                    command = f'ffmpeg -i "{in_path}" -crf {self.quality} -vf "drawtext=fontfile=Arial.ttf:text=\'%{{frame_num}}\':start_number=0:x=(w-tw)/2:y=h-(2*lh):fontcolor=black:fontsize=20:box=1:boxcolor=white:boxborderw=5" -c:a copy -y "{out_path}" -hide_banner -loglevel error'
                subprocess.check_output(command, shell=True)
                subprocess.call(command, shell=True, stdout=subprocess.PIPE)
            except:
                simba_cw = os.path.dirname(simba.__file__)
                simba_font_path = pathlib.Path(
                    simba_cw, "assets", "UbuntuMono-Regular.ttf"
                )
                if self.gpu:
                    command = f'ffmpeg -hwaccel auto -c:v h264_cuvid -i "{in_path}" -vf "drawtext=fontsize=24:fontfile={simba_font_path}:text=%{{n}}:x=(w-tw)/2:y=h-th-10:fontcolor=white:box=1:boxcolor=white@0.5" -c:v h264_nvenc -preset {self.quality} -c:a copy "{out_path}" -y -hide_banner -loglevel error'
                else:
                    command = f'ffmpeg -i "{in_path}" -vf "drawtext=fontfile={simba_font_path}:text=\'%{{frame_num}}\':start_number=1:x=(w-tw)/2:y=h-(2*lh):fontcolor=black:fontsize=20:box=1:boxcolor=white:boxborderw=5" -crf {self.quality} -c:a copy -y "{out_path}" -hide_banner -loglevel error'
                subprocess.call(command, shell=True, stdout=subprocess.PIPE)
        self.replace_files_in_temp()
        print("Applying frame count complete...")

    def apply_clahe(self):
        self.videos_to_frm_cnt = self.find_relevant_videos(variable="clahe")
        self.create_process_dir()
        for video, video_data in self.videos_to_frm_cnt.items():
            clahe_filter = cv2.createCLAHE(clipLimit=2, tileGridSize=(16, 16))
            print(
                f"Applying CLAHE {video}... (note: process can be slow for long videos)"
            )
            in_path, out_path = video_data["path"], os.path.join(
                self.process_dir, os.path.basename(video_data["path"])
            )
            video_info = get_video_meta_data(in_path)
            cap = cv2.VideoCapture(in_path)
            fps, width, height = (
                video_info["fps"],
                video_info["width"],
                video_info["height"],
            )
            writer = cv2.VideoWriter(
                out_path, cv2.VideoWriter_fourcc(*"MP4V"), fps, (width, height), 0
            )
            while True:
                ret, image = cap.read()
                if ret == True:
                    img = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
                    clahe_img = clahe_filter.apply(img)
                    writer.write(clahe_img)
                else:
                    break
            cap.release()
            writer.release()
        self.replace_files_in_temp()
        print("Applying CLAHE complete...")

    def crop_videos(self):
        self.videos_to_crop = self.find_relevant_videos(variable="crop")
        self.create_process_dir()
        for video, video_data in self.videos_to_crop.items():
            print(f"Applying crop {video}...")
            if video_data["last_operation"] == "crop":
                self.quality = video_data["output_quality"]
            in_path, out_path = video_data["path"], os.path.join(
                self.process_dir, os.path.basename(video_data["path"])
            )
            crop_settings = self.video_dict["video_data"][video]["crop_settings"]
            width, height = str(crop_settings["width"]), str(crop_settings["height"])
            top_left_x, top_left_y = str(crop_settings["top_left_x"]), str(
                crop_settings["top_left_y"]
            )
            if self.gpu:
                command = f'ffmpeg -hwaccel auto -c:v h264_cuvid -i "{in_path}" -vf "crop={width}:{height}:{top_left_x}:{top_left_y}" -c:v h264_nvenc -preset {self.quality} -c:a copy "{out_path}" -hide_banner -loglevel error'
            else:
                command = f'ffmpeg -i "{in_path}" -vf "crop={width}:{height}:{top_left_x}:{top_left_y}" -c:v libx264 -crf {self.quality} -c:a copy "{out_path}" -hide_banner -loglevel error'
            subprocess.call(command, shell=True, stdout=subprocess.PIPE)
        self.replace_files_in_temp()
        print("Applying crop complete...")

    def copy_videos_to_temp_dir(self):
        for video, video_data in self.video_dict["video_data"].items():
            source = video_data["video_info"]["file_path"]
            print("Making a copy of {} ...".format(os.path.basename(source)))
            destination = os.path.join(self.temp_dir, os.path.basename(source))
            shutil.copyfile(source, destination)

    def move_all_processed_files_to_output_folder(self):
        final_videos_path_lst = [
            f for f in glob.glob(self.temp_dir + "/*") if os.path.isfile(f)
        ]
        for file_path in final_videos_path_lst:
            shutil.copy(
                file_path, os.path.join(self.out_dir, os.path.basename(file_path))
            )
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)
        if os.path.exists(self.process_dir):
            shutil.rmtree(self.process_dir)


# test = FFMPEGCommandCreator(json_path='/Users/simon/Desktop/train_model_project/project_folder/videos_2/batch_process.json_log')
# test.crop_videos()
# test.clip_videos()
# test.downsample_videos()
# test.apply_fps()
# test.apply_grayscale()
# test.apply_frame_count()
# test.apply_clahe()
# test.move_all_processed_files_to_output_folder()



__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import List, Optional, Tuple

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_if_filepath_list_is_empty, check_int,
                                check_that_column_exist)
from simba.utils.data import detect_bouts
from simba.utils.enums import Formats
from simba.utils.errors import NoFilesFoundError, NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, get_video_meta_data, read_df
from simba.utils.warnings import NoDataFoundWarning


class ClassifierValidationClips(ConfigReader):
    """
    Create video clips with overlaid classified events for detection of false positive event bouts.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter int window: Number of seconds before and after the event bout that should be included in the output video.
    :parameter str clf_name: Name of the classifier to create validation videos for.
    :parameter bool clips: If True, creates individual video file clips for each validation bout.
    :parameter Tuple[int, int, int] text_clr: Color of text overlay in BGR.
    :parameter Optional[Tuple[int, int, int]] highlight_clr: Color of text when probability values are above threshold. If None, same as text_clr.
    :parameter float video_speed:  FPS rate in relation to original video. E.g., the same as original video if 1.0. Default: 1.0.
    :parameter bool concat_video:  If True, creates a single video including all events bouts for each video. Default: False.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/classifier_validation.md#classifier-validation>`_.

    Examples
    ----------
    >>> _ = ClassifierValidationClips(config_path='MyProjectConfigPath', window=5, clf_name='Attack', text_clr=(255,255,0), clips=False, concat_video=True).run()
    """

    def __init__(
        self,
        config_path: str,
        window: int,
        clf_name: str,
        clips: bool,
        data_paths: List[str],
        text_clr: Tuple[int, int, int],
        concat_video: bool = False,
        video_speed: float = 1.0,
        highlight_clr: Optional[Tuple[int, int, int]] = None,
    ):
        super().__init__(config_path=config_path)
        if (not clips) and (not concat_video):
            raise NoSpecifiedOutputError(
                msg="Please select to create clips and/or a concatenated video"
            )

        check_int(name="Time window", value=window)
        self.window, self.clf_name = int(window), clf_name
        self.clips, self.concat_video, self.video_speed, self.highlight_clr = (
            clips,
            concat_video,
            video_speed,
            highlight_clr,
        )
        self.p_col = f"Probability_{self.clf_name}"
        self.text_clr, self.data_paths = text_clr, data_paths
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.font = cv2.FONT_HERSHEY_SIMPLEX
        if not os.path.exists(self.clf_validation_dir):
            os.makedirs(self.clf_validation_dir)
        check_if_filepath_list_is_empty(
            filepaths=self.data_paths,
            error_msg="No data found in the project_folder/csv/machine_results directory",
        )
        print(f"Processing {str(len(self.data_paths))} files...")

    def __insert_inter_frms(self):
        """
        Helper to create N blank frames separating the classified event bouts.
        """

        for i in range(int(self.fps)):
            inter_frm = np.full(
                (int(self.video_info["height"]), int(self.video_info["width"]), 3),
                (49, 32, 189),
            ).astype(np.uint8)
            cv2.putText(
                inter_frm,
                "Bout #{}".format(str(self.bout_cnt + 1)),
                (
                    10,
                    (self.video_info["height"] - self.video_info["height"])
                    + self.spacing_scale,
                ),
                self.font,
                self.font_size,
                (0, 0, 0),
                2,
            )
            self.concat_writer.write(inter_frm)

    def run(self):
        """
        Method to generate clips. Results are saved in the ``project_folder/frames/output/classifier_validation directory``
        directory of the SimBA project.

        Returns
        -------
        None
        """
        for file_cnt, file_path in enumerate(self.data_paths):
            self.data_df = read_df(file_path, self.file_type)
            check_that_column_exist(
                df=self.data_df, column_name=self.p_col, file_name=file_path
            )
            _, file_name, _ = get_fn_ext(file_path)
            self.video_path = self.find_video_of_file(
                video_dir=self.video_dir, filename=file_name
            )
            if not self.video_path:
                raise NoFilesFoundError(
                    msg=f"Could not find a video file representing {file_name} in the project_folder/videos directory"
                )
            self.video_info = get_video_meta_data(video_path=self.video_path)
            self.fps = int(self.video_info["fps"])
            self.video_fps = int(self.fps * self.video_speed)
            if self.video_fps < 1:
                self.video_fps = 1
            self.space_scale, self.radius_scale, self.res_scale, self.font_scale = (
                60,
                12,
                1500,
                1.5,
            )
            self.max_dim = max(self.video_info["width"], self.video_info["height"])
            self.circle_scale = int(self.radius_scale / (self.res_scale / self.max_dim))
            self.font_size = float(self.font_scale / (self.res_scale / self.max_dim))
            self.spacing_scale = int(self.space_scale / (self.res_scale / self.max_dim))
            cap = cv2.VideoCapture(self.video_path)
            clf_bouts = detect_bouts(
                data_df=self.data_df, target_lst=[self.clf_name], fps=self.fps
            ).reset_index(drop=True)
            if self.concat_video:
                self.concat_video_save_path = os.path.join(
                    self.clf_validation_dir,
                    self.clf_name + "_" + file_name + "_all_events.mp4",
                )
                self.concat_writer = cv2.VideoWriter(
                    self.concat_video_save_path,
                    self.fourcc,
                    self.video_fps,
                    (int(self.video_info["width"]), int(self.video_info["height"])),
                )
                self.bout_cnt = 0
                self.__insert_inter_frms()
            if len(clf_bouts) == 0:
                NoDataFoundWarning(
                    msg=f"Skipping video {file_name}: No classified behavior detected..."
                )
                continue
            for bout_cnt, bout in clf_bouts.iterrows():
                clip_timer = SimbaTimer(start=True)
                self.bout_cnt = bout_cnt
                event_start_frm, event_end_frm = bout["Start_frame"], bout["End_frame"]
                start_window = int(
                    event_start_frm - (int(self.video_info["fps"]) * self.window)
                )
                end_window = int(
                    event_end_frm + (int(self.video_info["fps"]) * self.window)
                )
                self.save_path = os.path.join(
                    self.clf_validation_dir,
                    self.clf_name + "_" + str(bout_cnt) + "_" + file_name + ".mp4",
                )
                if start_window < 0:
                    start_window = 0
                current_frm = deepcopy(start_window)
                if end_window > len(self.data_df):
                    end_window = len(self.data_df)
                if self.clips:
                    bout_writer = cv2.VideoWriter(
                        self.save_path,
                        self.fourcc,
                        self.video_fps,
                        (int(self.video_info["width"]), int(self.video_info["height"])),
                    )
                event_frm_count = end_window - start_window
                print(start_window, end_window, current_frm)
                frm_cnt = 0
                cap.set(1, current_frm)
                while current_frm < end_window:
                    ret, img = cap.read()
                    p, clf_val = round(
                        float(self.data_df.loc[current_frm, self.p_col]), 3
                    ), int(self.data_df.loc[current_frm, self.clf_name])
                    self.add_spacer = 2
                    cv2.putText(
                        img,
                        "{} event # {}".format(self.clf_name, str(bout_cnt + 1)),
                        (
                            10,
                            (self.video_info["height"] - self.video_info["height"])
                            + self.spacing_scale * self.add_spacer,
                        ),
                        self.font,
                        self.font_size,
                        self.text_clr,
                        2,
                    )
                    self.add_spacer += 1
                    cv2.putText(
                        img,
                        "Total frames of event: {}".format(str(event_frm_count)),
                        (
                            10,
                            (self.video_info["height"] - self.video_info["height"])
                            + self.spacing_scale * self.add_spacer,
                        ),
                        self.font,
                        self.font_size,
                        self.text_clr,
                        2,
                    )
                    self.add_spacer += 1
                    cv2.putText(
                        img,
                        "Frames of event {} to {}".format(
                            str(start_window), str(end_window)
                        ),
                        (
                            10,
                            (self.video_info["height"] - self.video_info["height"])
                            + self.spacing_scale * self.add_spacer,
                        ),
                        self.font,
                        self.font_size,
                        self.text_clr,
                        2,
                    )
                    self.add_spacer += 1
                    cv2.putText(
                        img,
                        "Frame number: {}".format(str(current_frm)),
                        (
                            10,
                            (self.video_info["height"] - self.video_info["height"])
                            + self.spacing_scale * self.add_spacer,
                        ),
                        self.font,
                        self.font_size,
                        self.text_clr,
                        2,
                    )
                    self.add_spacer += 1
                    if (self.highlight_clr != None) and (clf_val == 1):
                        cv2.putText(
                            img,
                            f"Frame {self.clf_name} probability: {p}",
                            (
                                10,
                                (self.video_info["height"] - self.video_info["height"])
                                + self.spacing_scale * self.add_spacer,
                            ),
                            self.font,
                            self.font_size,
                            self.highlight_clr,
                            2,
                        )
                    else:
                        cv2.putText(
                            img,
                            f"Frame {self.clf_name} probability: {p}",
                            (
                                10,
                                (self.video_info["height"] - self.video_info["height"])
                                + self.spacing_scale * self.add_spacer,
                            ),
                            self.font,
                            self.font_size,
                            self.text_clr,
                            2,
                        )
                    print(
                        f"Frame {str(frm_cnt)} / {str(event_frm_count)}, Event {str(bout_cnt+1)}/{str(len(clf_bouts))}, Video {str(file_cnt+1)}/{str(len(self.machine_results_paths))}..."
                    )
                    if self.clips:
                        bout_writer.write(img)
                    if self.concat_video:
                        self.concat_writer.write(img)
                    current_frm += 1
                    frm_cnt += 1
                if self.clips:
                    bout_writer.release()
                if self.concat_video and bout_cnt != len(clf_bouts) - 1:
                    self.__insert_inter_frms()
            if self.concat_video:
                self.concat_writer.release()
        self.timer.stop_timer()
        stdout_success(
            msg="All validation clips complete. Files are saved in the project_folder/frames/output/classifier_validation directory of the SimBA project",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = ClassifierValidationClips(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                  window=10,
#                                  clf_name='Attack',
#                                  clips=False,
#                                  concat_video=True,
#                                  highlight_clr=(255, 0, 0),
#                                  video_speed=0.5,
#                                  text_clr=(0, 0, 255))
# test.create_clips()

# test = ClassifierValidationClips(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini',
#                                  window=1,
#                                  clf_name='Attack',
#                                  clips=False,
#                                  concat_video=True,
#                                  text_clr=(0, 0, 255))
# test.create_clips()



__author__ = "Simon Nilsson"

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader


class Interpolate(ConfigReader):
    """
    Interpolate missing body-parts in pose-estimation data.

    Parameters
    ----------
    config_file_path: str
        path to SimBA project config file in Configparser format
    in_file: pd.DataFrame
        Pose-estimation data

    Notes
    -----
    `Interpolation tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#to-import-multiple-dlc-csv-files>`__.

    Examples
    -----
    >>> body_part_interpolator = Interpolate(config_file_path='MyProjectConfig', in_file=input_df)
    >>> body_part_interpolator.detect_headers()
    >>> body_part_interpolator.fix_missing_values(method_str='Body-parts: Nearest')
    >>> body_part_interpolator.reorganize_headers()

    """

    def __init__(self, config_file_path: str, in_file: pd.DataFrame):
        super().__init__(config_path=config_file_path, read_video_info=False)
        self.in_df = in_file

    def detect_headers(self):
        """
        Method to detect multi-index headers and set values to numeric in input dataframe
        """
        self.multi_index_headers_list = []
        self.in_df.columns = self.bp_headers
        self.header_col_cnt = self.get_number_of_header_columns_in_df(df=self.in_df)
        self.current_df = (
            self.in_df.iloc[self.header_col_cnt :]
            .apply(pd.to_numeric)
            .reset_index(drop=True)
        )
        self.multi_index_headers = self.in_df.iloc[: self.header_col_cnt]
        if self.header_col_cnt == 2:
            self.idx_names = ["scorer", "bodyparts", "coords"]
            for column in self.multi_index_headers:
                self.multi_index_headers_list.append(
                    (
                        column,
                        self.multi_index_headers[column][0],
                        self.multi_index_headers[column][1],
                    )
                )
        else:
            self.idx_names = ["scorer", "individuals", "bodyparts", "coords"]
            for column in self.multi_index_headers:
                self.multi_index_headers_list.append(
                    (
                        column,
                        self.multi_index_headers[column][0],
                        self.multi_index_headers[column][1],
                        self.multi_index_headers[column][2],
                    )
                )

    def fix_missing_values(self, method_str: str):
        """
        Method to interpolate missing values in pose-estimation data.

        Parameters
        ----------
        method_str: str
            String representing interpolation method. OPTIONS: 'None','Animal(s): Nearest', 'Animal(s): Linear',
            'Animal(s): Quadratic','Body-parts: Nearest', 'Body-parts: Linear', 'Body-parts: Quadratic'

        """

        interpolation_type, interpolation_method = (
            method_str.split(":")[0],
            method_str.split(":")[1].replace(" ", "").lower(),
        )
        self.animal_df_list, self.header_list_p = [], []
        if interpolation_type == "Animal(s)":
            for animal in self.multi_animal_id_list:
                currentAnimalX, currentAnimalY, currentAnimalP = (
                    self.animal_bp_dict[animal]["X_bps"],
                    self.animal_bp_dict[animal]["Y_bps"],
                    self.animal_bp_dict[animal]["P_bps"],
                )
                header_list_xy = []
                for (
                    col1,
                    col2,
                    col3,
                ) in zip(currentAnimalX, currentAnimalY, currentAnimalP):
                    header_list_xy.extend((col1, col2))
                    self.header_list_p.append(col3)
                self.animal_df_list.append(self.current_df[header_list_xy])
            for loop_val, animal_df in enumerate(self.animal_df_list):
                repeat_bol = animal_df.eq(animal_df.iloc[:, 0], axis=0).all(
                    axis="columns"
                )
                indices_to_replace_animal = repeat_bol.index[repeat_bol].tolist()
                print(
                    f"Detected {str(len(indices_to_replace_animal))} missing pose-estimation frames for {str(self.multi_animal_id_list[loop_val])}..."
                )
                animal_df.loc[indices_to_replace_animal] = np.nan
                self.animal_df_list[loop_val] = (
                    animal_df.interpolate(method=interpolation_method, axis=0)
                    .ffill()
                    .bfill()
                )
            self.new_df = pd.concat(self.animal_df_list, axis=1).fillna(0)

        if interpolation_type == "Body-parts":
            for animal in self.animal_bp_dict:
                for x_bps_name, y_bps_name in zip(
                    self.animal_bp_dict[animal]["X_bps"],
                    self.animal_bp_dict[animal]["Y_bps"],
                ):
                    zero_indices = self.current_df[
                        (self.current_df[x_bps_name] == 0)
                        & (self.current_df[y_bps_name] == 0)
                    ].index.tolist()
                    self.current_df.loc[zero_indices, [x_bps_name, y_bps_name]] = np.nan
                    self.current_df[x_bps_name] = (
                        self.current_df[x_bps_name]
                        .interpolate(method=interpolation_method, axis=0)
                        .ffill()
                        .bfill()
                    )
                    self.current_df[y_bps_name] = (
                        self.current_df[y_bps_name]
                        .interpolate(method=interpolation_method, axis=0)
                        .ffill()
                        .bfill()
                    )
            self.new_df = self.current_df.fillna(0)

    def reorganize_headers(self):
        """
        Method to re-insert original multi-index headers
        """
        loop_val = 2
        for p_col_name in self.header_list_p:
            p_col = list(self.in_df[p_col_name].iloc[self.header_col_cnt :])
            self.new_df.insert(loc=loop_val, column=p_col_name, value=p_col)
            loop_val += 3
        self.new_df.columns = pd.MultiIndex.from_tuples(
            self.multi_index_headers_list, names=self.idx_names
        )


# config_file_path = r"/Users/simon/Desktop/envs/troubleshooting/Vince_time_bins/project_folder/project_config.ini"
# in_file = r"/Users/simon/Desktop/envs/troubleshooting/Vince_time_bins/project_folder/csv/input_csv/2022-06-15_15-26-48_ArC_H_1_1Hrcropped.csv"
# interpolation_method = 'Body-parts: Linear'
# csv_df = pd.read_csv(in_file, index_col=0)
# interpolate_body_parts = Interpolate(config_file_path, csv_df)
# interpolate_body_parts.detect_headers()
# interpolate_body_parts.fix_missing_values(interpolation_method)
# interpolate_body_parts.reorganize_headers()



__author__ = "Simon Nilsson"

import os
from typing import Union

import numpy as np
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from copy import deepcopy

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.enums import TagNames
from simba.utils.errors import InvalidInputError, NoFilesFoundError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df, write_df)


class OutlierCorrecterMovementAdvanced(ConfigReader, FeatureExtractionMixin):
    """
    Performs outlier correction that allows different heuristic outlier criteria for different animals or body-parts.
    For example, correct some outlier body-parts with a movement heuristic criteria of 2x above the mean movement,
    and other body-parts with a heuristic critera of 1.5x above the mean movement.

    .. note::
       See `notebook <https://simba-uw-tf-dev.readthedocs.io/en/latest/nb/advanced_outlier_correction.html>`__. for example use-case.

    .. image:: _static/img/movement_outlier.png
       :width: 400
       :align: center

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter str input_dir: Directory containing input files. E.g., `project_folder/csv/input_csv` directory.
    :parameter Literal['animal', 'body-part'] type: If the rules are defined on animal or body-part level. E.g., If one heuristic rule per animal then `animal`. If one heurstic rule prr body-part then `body-parts`
    :parameter Literal['mean', 'median'] agg_method: If to use the mean or median to compute criterion.
    :parameter dict criterion_body_parts: The body-parts used to calculate the size of the animals.
    :parameter dict settings: The heuristic multiplier rules to use for each body part or animal.

    :example:
    >>> settings = {'Simon': 1.1, 'JJ': 1.2}
    >>> criterion_body_parts = {'Simon': ['Nose_1', 'Tail_base_1'], 'JJ': ['Nose_2', 'Tail_base_2']}
    >>> outlier_corrector = OutlierCorrecterMovementAdvanced(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', input_dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/input_csv', type='animal', agg_method='mean', criterion_body_parts=criterion_body_parts, settings=settings)
    >>> outlier_corrector.run()
    >>> settings = {'Simon': {'Ear_left_1': 1.1, 'Ear_right_1': 5.1, 'Nose_1': 2.1, 'Center_1': 1.5, 'Lat_left_1': 3.1, 'Lat_right_1': 1.9, 'Tail_base_1': 2.3, 'Tail_end_1': 1.4}, 'JJ': {'Ear_left_2': 1.2, 'Ear_right_2': 1.2, 'Nose_2': 2, 'Center_2': 4.1, 'Lat_left_2': 9, 'Lat_right_2': 1.2, 'Tail_base_2': 1.6, 'Tail_end_2': 2.2}}
    >>> outlier_corrector = OutlierCorrecterMovementAdvanced(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', input_dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/input_csv', type='body-part', agg_method='mean', criterion_body_parts=criterion_body_parts, settings=settings)
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        input_dir: Union[str, os.PathLike],
        criterion_body_parts: dict,
        type: Literal["animal", "body-part"],
        agg_method: Literal["mean", "median"],
        settings: dict,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        if type not in ["animal", "body-parts"]:
            raise InvalidInputError(
                msg=f"Type {type} not supported. Valid options: {['animal', 'body-parts']}"
            )
        if agg_method not in ["mean", "median"]:
            raise InvalidInputError(
                msg=f"Aggregation method {agg_method} not supported. Valid options: {['mean', 'median']}"
            )
        self.data_files = find_files_of_filetypes_in_directory(
            directory=input_dir, extensions=["." + self.file_type]
        )
        if len(self.data_files) == 0:
            raise NoFilesFoundError(
                msg=f"No data of filetype {input_dir} for in directory {self.file_type}",
                source=self.__class__.__name__,
            )
        self.settings, self.type, self.agg_method = settings, type, agg_method
        self.save_path, self.criterion_body_parts = (
            self.outlier_corrected_movement_dir,
            criterion_body_parts,
        )

    def fix_settings_for_animal_input(self):
        new_settings = {}
        for animal_name, animal_bps in self.animal_bp_dict.items():
            if animal_name in self.settings.keys():
                if animal_name not in new_settings.keys():
                    new_settings[animal_name] = {}
                for body_part in animal_bps["X_bps"]:
                    new_settings[animal_name][body_part[:-2]] = self.settings[
                        animal_name
                    ]
        self.settings = new_settings

    def _save_data_log(self, data_log: dict):
        out = pd.DataFrame(
            columns=[
                "VIDEO",
                "ANIMAL",
                "BODY-PART",
                "OUTLIER COUNT",
                "OUTLIER RATIO (%)",
            ]
        )
        save_path = os.path.join(
            self.logs_path, f"movement_outliers_{self.datetime}.csv"
        )
        for video_name, video_data in data_log.items():
            for animal_name, animal_data in video_data.items():
                for body_part_name, body_part_data in animal_data.items():
                    out.loc[len(out)] = [
                        video_name,
                        animal_name,
                        body_part_name,
                        body_part_data["outlier_cnt"],
                        body_part_data["outlier_ratio"],
                    ]
        write_df(df=out, file_type="csv", save_path=save_path)
        stdout_success(
            msg=f"Movement outlier log saved at {save_path}",
            source=self.__class__.__name__,
        )

    def run(self):
        data_log = {}
        if self.type == "animal":
            self.fix_settings_for_animal_input()
        for file_cnt, file_path in enumerate(self.data_files):
            video_timer, video_log = SimbaTimer(start=True), {}
            _, self.video_name, _ = get_fn_ext(file_path)
            print(
                "Processing video {}. Video {}/{}...".format(
                    self.video_name, str(file_cnt + 1), str(len(self.data_files))
                )
            )
            self.data_df = read_df(file_path, self.file_type, check_multiindex=True)
            self.data_df = self.insert_column_headers_for_outlier_correction(
                data_df=self.data_df, new_headers=self.bp_headers, filepath=file_path
            )
            self.results = deepcopy(self.data_df)
            self.data_df_combined = self.create_shifted_df(df=self.data_df)
            animal_movement_agg = {}
            self.movements = pd.DataFrame()
            for animal_name, animal_bps in self.animal_bp_dict.items():
                if animal_name in self.criterion_body_parts.keys():
                    animal_bp_headers = np.array(
                        [
                            item
                            for pair in zip(animal_bps["X_bps"], animal_bps["Y_bps"])
                            for item in pair
                        ]
                    ).reshape(len(animal_bps["X_bps"]), 2)
                    animal_criterion_bps = self.criterion_body_parts[animal_name]
                    bp_1_headers = [
                        animal_criterion_bps[0] + "_x",
                        animal_criterion_bps[0] + "_y",
                    ]
                    bp_2_headers = [
                        animal_criterion_bps[1] + "_x",
                        animal_criterion_bps[1] + "_y",
                    ]
                    print(self.data_df.columns)
                    distances = self.framewise_euclidean_distance(
                        location_1=self.data_df[bp_1_headers].values,
                        location_2=self.data_df[bp_2_headers].values,
                        px_per_mm=1,
                    )
                    if self.agg_method == "mean":
                        animal_movement_agg[animal_name] = np.mean(distances).astype(
                            int
                        )
                    if self.agg_method == "median":
                        animal_movement_agg[animal_name] = np.median(distances).astype(
                            int
                        )
                    for bps in animal_bp_headers:
                        bp_name = bps[0][:-2]
                        shifted_bps = [bps[0] + "_shifted", bps[1] + "_shifted"]
                        distances = self.framewise_euclidean_distance(
                            location_1=self.data_df_combined[bps].values,
                            location_2=self.data_df_combined[shifted_bps].values,
                            px_per_mm=1,
                        )
                        self.movements[bp_name] = distances

            for animal_name, animal_bps in self.settings.items():
                video_log[animal_name] = {}
                for bp, criterion_multiplier in animal_bps.items():
                    if animal_name in animal_movement_agg.keys():
                        bp_criterion = (
                            animal_movement_agg[animal_name] * criterion_multiplier
                        )
                        over_criterion_idx = list(
                            self.movements.index[self.movements[bp] > bp_criterion]
                        )
                        video_log[animal_name][bp] = {
                            "outlier_cnt": len(over_criterion_idx),
                            "outlier_ratio": len(over_criterion_idx)
                            / len(self.movements),
                        }
                        self.results.loc[over_criterion_idx, [bp + "_x", bp + "_y"]] = (
                            np.nan
                        )
                        self.results[[bp + "_x", bp + "_y"]] = (
                            self.results[[bp + "_x", bp + "_y"]]
                            .ffill()
                            .bfill()
                            .astype(int)
                        )

            df_save_path = os.path.join(
                self.outlier_corrected_movement_dir,
                self.video_name + f".{self.file_type}",
            )
            write_df(df=self.results, file_type=self.file_type, save_path=df_save_path)
            video_timer.stop_timer()
            stdout_success(
                msg=f"Movement outliers complete for video {self.video_name}.",
                elapsed_time=video_timer.elapsed_time_str,
                source=self.__class__.__name__,
            )
            data_log[self.video_name] = video_log
        self._save_data_log(data_log=data_log)
        stdout_success(
            msg=f"{len(self.data_files)} video(s) corrected for movement outliers. Saved in {self.outlier_corrected_movement_dir}",
            source=self.__class__.__name__,
        )


#
# settings = {'Simon': 2.5} #'JJ': 1.2
# # settings = {'Simon': {'Ear_left_1': 1.1,
# #                       'Ear_right_1': 5.1,
# #                       'Nose_1': 2.1,
# #                       'Center_1': 1.5,
# #                       'Lat_left_1': 3.1,
# #                       'Lat_right_1': 1.9,
# #                       'Tail_base_1': 2.3},
# #                       #'Tail_end_1': 1.4},
# #                'JJ': {'Ear_left_2': 1.2,
# #                       'Ear_right_2': 1.2,
# #                       'Nose_2': 2,
# #                       'Center_2': 4.1,
# #                       'Lat_left_2': 9,
# #                       'Lat_right_2': 1.2,
# #                       'Tail_base_2': 1.6,
# #                       'Tail_end_2': 2.2}}
# criterion_body_parts = {'Simon': ['Nose_1', 'Tail_base_1']} #'JJ': ['Nose_2', 'Tail_base_2']
#
#
# test = OutlierCorrecterMovementAdvanced(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                         input_dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/input_csv',
#                                         criterion_body_parts=criterion_body_parts,
#                                         type='animal',
#                                         agg_method='mean',
#                                         settings=settings)
# test.run()

#
